var AWS = require('hws-sdk/lib/core');
var fs = require('fs');
var commonUtils = require('./common/CommonUtils.js');
var crypto = require('crypto');
function DownloadPool(){

}

constructor: function DownloadPool(userId,s3,downloadConfig,db) {
    var self=this;
    AWS.SequentialExecutor.call(self);
    this.userId=userId;
    this.s3=s3;
    this.db=db;

    if(downloadConfig){
        if(downloadConfig.isMultiDownload){
            this.isMultiDownload = downloadConfig.isMultiDownload;
        }else{
            this.isMultiDownload = false;
        }
        if(downloadConfig.multiDownloadNum){
            this.multiDownloadNum = downloadConfig.multiDownloadNum;
        }else{
            this.multiDownloadNum = 5;
        }

        if(downloadConfig.maxTaskNum){
            this.maxTaskNum = downloadConfig.maxTaskNum;
        }else{
            this.maxTaskNum = 20;
        }
    }
    else{
        this.isMultiDownload = false;
        this.multiDownloadNum = 5;
        this.maxTaskNum = 20;
    }

    this.pauseAll=false;
    this.abortAll=false;

    this.runningMap={};//运行集合
    this.runningSize=0;//运行数
};



//读取并执行任务
DownloadPool.prototype.run=function run(){
    var self=this;
    if(this.maxTaskNum > this.runningSize && this.waitSize>0 && !this.pauseAll){
        self.waitSize--;
        self.runningSize++;

        var queryString={userId:self.userId,stat:"WAIT",type:"getObj"};
        self.db.getAndModify(queryString,function(err,taskRecord){
            if(!err && taskRecord) {
                var task = taskRecord.value;
                if (task) {
                    task.startTime = new Date();
                    global.logger.debug("execute task:",task.bucket," ",task.key," ",task.type);
                    task.stat = "RUNNING";
                    self.runningMap[task.id] = task;
                    self.emit("onStart", [task]);
                        var filePath = task.path;
                        var fatherPath = filePath.substring(0, filePath.lastIndexOf("/"));
                        fs.exists(fatherPath, function (exists) {
                            if (!exists) {
                                self.createLocalFolder(fatherPath, function (err) {
                                    if(err){
                                        global.logger.error('createLocalFolder error',err);
                                        task.stat = "FAILED";
                                        task.errMessage = err.code;
                                        self.finish(task);
                                    }else{
                                        self.download(task);
                                    }
                                });
                            }
                            else {
                                self.download(task);
                            }
                        });
                    }
                else {
                    self.waitSize++;
                    self.runningSize--;
                }
            }
            else{
                self.waitSize++;
                self.runningSize--;
                }
        });
    }
};

//任务结束，包含正常结束、暂停、失败、取消
DownloadPool.prototype.finish=function finish(task,callback){
    var self=this;
    if(this.runningMap[task.id]){
        this.runningSize--;
        delete this.runningMap[task.id];
        var times = 0;
        if(task.startTime){
            times = new Date().getTime() - task.startTime.getTime();
        }
        if(task.stat==="RUNNING"){
            var stat = fs.lstatSync(task.path);
            var isDirectory = stat.isDirectory();
            if(global.isMD5check === "Y"&& !isDirectory){
                var params = {//请求对象自定义数据中的文件MD5
                    Bucket : task.bucket,
                    Key:  task.key
                };
                s3.headObject(params,function (err,data) {
                    var MD5 = data.Metadata.md5chksum;
                    if(MD5 == undefined){
                        task.stat = "SUCCESS";
                        delete task['active'];
                        delete task["download"];
                        delete task['req'];
                        task.endTime = new Date();
                        self.db.updateStat(task,function(err,data){
                            if(!err) {
                                self.emit("end",[task]);
                            }
                            else{
                                global.logger.error(err)
                            }
                        });
                    }else{
                        task.stat = "MD5Check";
                        delete task['speed'];
                        delete task['spTime'];
                        var md5num = crypto.createHash('md5');
                        var chunk =fs.createReadStream(task.path).on('data',function(chunks){
                            md5num.update(chunks);
                        }).on('end',function () {
                            var fileMD5 = md5num.digest('hex');
                            if(MD5 = fileMD5){
                                task.stat = "MD5SUCCESS";
                                delete task['active'];
                                delete task["download"];
                                delete task['req'];
                                task.endTime = new Date();
                                self.db.updateStat(task,function(err,data){
                                    if(!err) {
                                        self.emit("end",[task]);
                                    }
                                    else{
                                        global.logger.error(err)
                                    }
                                });
                            }else{
                                global.logger.info("11111111")
                            }
                            chunk.destroy();
                        })
                    }
                });
                global.successLogger.info("download",'key:',task.key,"size:",task.size,"accountName:",global.currentUserName,"duration:",times,"status:","Success");
            }else{
                task.stat = "SUCCESS";
                delete task['active'];
                delete task["download"];
                delete task['req'];
                task.endTime = new Date();
                self.db.updateStat(task,function(err,data){
                    if(!err) {
                        self.emit("end",[task]);
                    }
                    else{
                        global.logger.error(err)
                    }
                });
            }
        }
        else if(task.stat==="PAUSE"){
            delete task['active'];
            delete task['speed'];
            delete task['spTime'];
            self.db.updateStat(task,function(err,data){
                if(!err) {
                    self.emit("onPause", [task]);
                }
                else{
                    global.logger.error(err)
                }
            });

        }
        else if(task.stat==="FAILED"){
            global.successLogger.info("download",'key:',task.key,"size:",task.size,"accountName:",global.currentUserName,"duration:",times,"errCode:",task.errMessage,"status:","Failed");
            delete task['speed'];
            delete task['spTime'];
            if(task.errMessage==='NetworkingError'){
                task.stat = 'WAIT';
                self.db.updateStat(task,function(err,data){
                    if(!err) {
                        self.emit("onWait", [task]);
                    }
                    else{
                        global.logger.error('[DB]--error',err);
                    }
                });
            }else{
                task.endTime = new Date();
                self.db.updateStat(task,function(err,data){
                    if(!err) {
                        global.logger.info('[FAILED]=====',err,data,task);
                        self.emit("failed", [task]);
                    }
                    else{
                        global.logger.error(err);
                    }
                });
            }
        }
        else if(task.stat==="ABORT"){
            self.db.delById(task.id,function(err,data){
                if(!err) {
                    self.emit("onAbort", [task]);
                }
                else{
                    global.logger.error(err)
                }
            });
        }
    }
    if(callback){
        callback.call();
    }
};



//下载文件
DownloadPool.prototype.download = function download(task){
    var self=this;
    if(task.path.length > 259 ){
        global.logger.error("The length of task path, and is:"+task.path.length);
        task.stat = "FAILED";
        task.errMessage = "ENOENT";
        self.finish(task);
        return false;
    }
    if(task.size===0 && task.path.charAt(task.path.length-1)=='/'){
        //文件大小为0
        fs.exists(task.path, function (exists) {
            if(exists){
                self.finish(task);
            }else{
                fs.mkdir(task.path,function(err){
                    if(err){
                        global.logger.error('mkdir error',err,task.path);
                        task.stat = "FAILED";
                        task.errMessage = err.code;
                        self.finish(task);
                    }else{
                        self.finish(task);
                    }
                });
            }
        });
    }
    else if(task.size <= 65536) {
        //文件大小为size<=64K
        var out = fs.createWriteStream(task.path)
            .on("open", function (fd) {
                if(fd) {
                    task.fd = fd;
                    var params = {Bucket: task.bucket, Key: task.key};
                    var req = self.s3.getObject(params);
                    if (req) {
                        var loaded = 0;
                        var readStream = req.createReadStream()
                            .on('data', function (chunk) {
                                loaded += chunk.length;
                                var info = {
                                    loaded: loaded,
                                    total: task.size
                                };
                                self.emit("progress", [task, info]);
                                chunk = null;
                                delete chunk;
                            })
                            .on('error', function (err) {
                                global.logger.error('[Download]',err);
                                delete task["req"];
                                task.stat = "FAILED";
                                task.errMessage = err.code;

                                fs.unlink(task.path,function(err){});
                                //用户取消任务
                                if (err.code === "RequestAbortedError") {
                                    task.stat = "ABORT";
                                }
                                out.end();
                            });
                        readStream.pipe(out);
                    }
                }
                else{
                    delete task["req"];
                    task.stat = "FAILED";
                    task.errMessage = "unknow err";
                    self.finish(task);
                }
            })
            .on("finish", function (fd) {
                self.finish(task);
            })
            .on('error', function (err) {
                global.logger.error(err);
                delete task["req"];
                task.stat = "FAILED";
                task.errMessage = err.code;

                self.finish(task);
            });
    }
    else{

        if(global.isMultiDownload){
            var params = {Bucket: task.bucket, Key: task.key, localpath : task.path, size : task.size};
            var options = {partSize: this.multiDownloadNum * 1024 * 1024,queueSize: 4};
            if(task.partSize){
                options.partSize=task.partSize;
            }
            if(task.queueSize){
                options.queueSize=task.queueSize;
            }
            if(task.completeMap){
                options.completeMap=task.completeMap;
            }
            if(task.totalDownloaded){
                options.totalDownloaded=task.totalDownloaded;
            }

            var download;
            global.logger.info("download:", params,options)
            download=this.s3.download(params,options).
                on('httpDownloadProgress', function (evt) {
                    self.emit("progress",[task,evt]);
                });

            if(download){
                task.download = download;
                download.send(function(err, data) {
                    if (err)
                    {
                        global.logger.error(err);
                        if(task.download){
                            if(task.download.partSize){
                                task.partSize=task.download.partSize;
                            }
                            if(task.download.completeMap){
                                task.completeMap=task.download.completeMap;
                            }
                            if(task.download.totalDownloadedBytes){
                                task.totalDownloaded=task.download.totalDownloadedBytes;
                            }
                            delete task["download"];
                            task.stat = "FAILED";
                            task.errMessage = err.code;
                            //用户取消任务
                            if(err.code==="RequestAbortedError") {
                                fs.unlink(task.path,function(err){});
                                task.stat = "ABORT";
                                if(task.completeMap){
                                    delete task.completeMap;
                                }
                            }
                            else if(err.code==="NetworkingError") {
                                global.logger.warn('[NetworkingError]',err);
                            }
                        }
                    }
                    else{
                        fs.rename(task.path, task.path,function(){});
                    }
                    self.finish(task);
                });
            }
        }
        else{
            var position = 0;
            var filePath = task.path;
            var exists = fs.existsSync(filePath);
            if(exists){
                var stat = fs.lstatSync(filePath);
                position=stat.size;
            }
            var out = fs.createWriteStream(filePath,{ flags: 'a+'})
                .on("open", function (fd) {
                    if(fd) {
                        var params = {Bucket: task.bucket, Key: task.key};
                        if(position>0){
                            params.Range = "bytes="+position+"-";
                        }
                        var req = self.s3.getObject(params);
                        task.req=req;
                        if (req) {
                            var loaded = position;
                            var readStream = req.createReadStream()
                                .on('data', function (chunk) {
                                    loaded += chunk.length;
                                    var info = {
                                        loaded: loaded,
                                        total: task.size
                                    };
                                    self.emit("progress", [task, info]);
                                    chunk = null;
                                    delete chunk;
                                })
                                .on('error', function (err) {
                                    global.logger.error(err);
                                    delete task["req"];
                                    task.stat = "FAILED";
                                    task.errMessage = err.code;
                                    //用户取消任务
                                    if (err.code === "RequestAbortedError") {
                                        task.stat = "ABORT";
                                        fs.unlink(task.path,function(err){});
                                    }
                                    out.end();
                                });
                            readStream.pipe(out);
                        }
                    }
                    else{
                        delete task["req"];
                        task.stat = "FAILED";
                        task.errMessage = "unknow err";
                        self.finish(task);
                    }
                })
                .on("finish", function (fd) {
                    if(task.stat==="RUNNING"){
                        fs.rename(task.path, task.path,function(){});
                    }
                    self.finish(task);
                })
                .on('error', function (err) {
                    global.logger.error(err);
                    delete task["req"];
                    task.stat = "FAILED";
                    task.errMessage = err.code;

                    self.finish(task);
                });
        }

    }

};


//创建本地文件系统文件夹
DownloadPool.prototype.createLocalFolder = function createLocalFolder(path,callback){

    var fatherPath=path.substring(0,path.lastIndexOf("/",path.length-2)+1);
    try{
        fs.exists(fatherPath,function(exists){
            if(exists){
                //创建本地文件系统文件夹
                fs.mkdir(path,function(err){
                    if(callback){
                        callback.call(err);
                    }
                });
            } else{
                //创建本地文件系统文件夹
                createLocalFolder(fatherPath,function(){
                    fs.mkdir(path,function(err){
                        if(callback){
                            callback.call(err);
                        }
                    });
                });

            }
        });
    }catch(e){
        global.logger.error('createLocalFolder error',e,path)
    }

};

DownloadPool.prototype.loadWaitTaskNum = function loadWaitTaskNum(callback){
    var self = this;
    var count = 0;
    self.db.count({userId:self.userId,stat:"WAIT",type:"getObj"},function(err,data){
        if(!err)
            self.waitSize=data;
    });
};
DownloadPool.prototype.loadRunningTaskNum = function loadRunningTaskNum(){
    var self = this;
    var count = 0;
    for(var key in self.runningMap){
        count++;
    }
    self.runningSize = count;
};


AWS.util.mixin(DownloadPool, AWS.SequentialExecutor);
module.exports = DownloadPool;