var AWS = require('hws-sdk/lib/core');
var fs = require('fs');
var path = require('path');
var db = require('./mongodb');
var DownloadPool = require("./DownloadPool");
var UploadBucketPool = require("./UploadBucketPool");
var commonUtils = require('./common/CommonUtils.js');
var collection;

function TaskManager(){

}

constructor: function TaskManager(userId,s3,taskNum,uploadConfig,downloadConfig) {
    var self=this;

    AWS.SequentialExecutor.call(self);
    this.userId=userId;
    this.s3=s3;

    if(uploadConfig){
        if(!uploadConfig.isMultiUpload){
            uploadConfig.isMultiUpload = false;
        }
        if(!uploadConfig.multiUploadNum){
            uploadConfig.multiUploadNum = 5;
        }
        if(!uploadConfig.maxUploadTasks){
            global.maxUploadTasks = 20;
        }
        if(!uploadConfig.maxTps){
            uploadConfig.maxTps = 15;
        }
    }
    else{
        uploadConfig.isMultiUpload = false;
        uploadConfig.multiUploadNum = 5;
        global.maxUploadTasks = 50;
        uploadConfig.maxTps = 15;
    }
    self.uploadConfig = uploadConfig;
    if(downloadConfig){
        if(!downloadConfig.isMultiDownload){
            downloadConfig.isMultiDownload = false;
        }
        if(!downloadConfig.multiDownloadNum){
            downloadConfig.multiDownloadNum = 5;
        }
        if(!downloadConfig.maxTaskNum){
            downloadConfig.maxTaskNum = 20;
        }
    }
    else{
        downloadConfig.isMultiDownload = false;
        downloadConfig.multiDownloadNum = 5;
        downloadConfig.maxTaskNum = 20;
    }
    self.downloadConfig = downloadConfig;
    this.pauseAll=false;
    this.abortAll=false;

    this.runningMap={};//所有任务运行集合
    this.uploadBucketPools = {};//上传池集合
    global.uploadRunningSize = 0;
    this.downloadPool = null;//下载池

    var self = this;
    db.getCollection(function(err,data){
        collection = data;
        if(collection){
            var queryString = {userId : self.userId, "stat" : "RUNNING"};
            var param = {$set: {stat : "PAUSE"}};
            //如果上次退出异常关闭，修正数据库中RUNNING==>>PAUSE
            db.updateProperties(queryString,param,function(err,data){
                //初始化下载池
                self.downloadPool = new DownloadPool(userId,s3,self.downloadConfig,db);
                self.downloadPool.on("progress", function(task, info){
                        self.emit("progress",[task, info]);
                    }).on("onStart", function (task) {
                        self.runningMap[task.id]=task;
                        self.emit("onStart",[task]);
                    }).
                    on("onWait", function (task) {
                        delete self.runningMap[task.id];
                        self.emit("onWait",[task]);
                    }).
                    on("onPause", function (task) {
                        delete self.runningMap[task.id];
                        self.emit("onPause",[task]);
                    }).
                    on("failed", function (task) {
                        delete self.runningMap[task.id];
                        self.emit("failed",[task]);
                    }).
                    on("onAbort", function (task) {
                        delete self.runningMap[task.id];
                        self.emit("onAbort",[task]);
                    }).
                    on("end", function (task) {
                        delete self.runningMap[task.id];
                        self.emit("end",[task]);
                    });
                self.emit("dbconn");//数据连接成功
            });

            self.time = setInterval(function(){
                if(self.downloadPool) {
                    self.downloadPool.run();
                }
            },213);
            self.time1 = setInterval(function(){
                if(self.uploadBucketPools) {
                    for (var key in self.uploadBucketPools) {
                        self.uploadBucketPools[key].run();
                    }
                }
            },217);
            self.loadWaitTasksItv = setInterval(function(){
                self.manageBucketPools();
                self.downloadPool.loadWaitTaskNum();
                self.downloadPool.loadRunningTaskNum();
                for(var key in self.uploadBucketPools){
                    self.uploadBucketPools[key].loadWaitTaskNum();
                }
            },3111);
        }
        if(err){
            global.logger.error('Collection Mongodb Err:',err);
        }
    });
};

//插入任务到数据库
TaskManager.prototype.push=function push(task){
    var self = this;
    if(task)
    {
        task.userId=self.userId;
        if(self.abortAll){
           return;
        }
        if(self.pauseAll && task.type !== 'recover'){
            task.stat="PAUSE";
            db.add(task,function(err,data){
                if(err){
                   global.logger.error(err);
                }
            });
        }
        else{
            //写入数据库
                db.add(task,function(err,data){
                    if(err){
                        global.logger.error(err);
                    }
                });
        }
    }
};




//任务结束，包含正常结束、暂停、失败、取消
TaskManager.prototype.finish=function finish(task,callback){
    var self=this;
    if(this.runningMap[task.id]) {
        delete this.runningMap[task.id];
        delete this.downloadPool.runningMap[task.id];
        if(this.uploadBucketPools[task.bucket] && this.uploadBucketPools[task.bucket].runningMap){
            global.uploadRunningSize = (global.uploadRunningSize-- > 0) ? global.uploadRunningSize : 0;//global.uploadRunningSize最小值为0
            delete this.uploadBucketPools[task.bucket].runningMap[task.id];
        }
        global.logger.debug('[task]--->finish-->',task);
        if (task.stat === "RUNNING") {
            task.stat = "SUCCESS";
            delete task['active'];
            delete task['speed'];
            delete task['spTime'];
            delete task['upload'];
            delete task['completeMap'];
            delete task['req'];
            task.endTime = new Date();
            self.db.updateStat(task,function(err,data){
                if(!err) {
                    self.emit("end",[task]);
                }
                else{
                    global.logger.error('[DB]--error',err);
                }
            });
        }
        else if (task.stat === "PAUSE") {
            delete task['active'];
            db.updateStat(task, function (err, data) {
                if (!err) {
                    self.emit("onPause", [task]);
                }
                else {
                    global.logger.error('[DB] error',err);
                }
            });

        }
        else if (task.stat === "FAILED") {
            db.updateStat(task, function (err, data) {
                if (!err) {
                    self.emit("failed", [task]);
                }
                else {
                    global.logger.error('[DB] error',err);
                }
            });

        }
        else if (task.stat === "ABORT") {
            db.delById(task.id, function (err, data) {
                if (!err) {
                    self.emit("onAbort", [task]);
                }
                else {
                    global.logger.error('[DB] error',err);
                }
            });
        }
    }
    if(callback){
        callback.call();
    }

};


//取消
TaskManager.prototype.abort = function abort(id){
    var self=this;
    var task=self.runningMap[id];
    if(task && task.stat==="RUNNING"){
        global.logger.debug("abort task:",task.bucket," ",task.key," ",task.type);
        //上传
        if(task.type==="putObj") {
            if (task.upload) {
                task.upload.abort();
            }
            else if(task.req){
                task.req.abort();
            }
        }
        //下载
        else if(task.type==="getObj"){
            //取消正在执行的req
            if(task.download)
            {
                task.download.abort();
                task.download=null;
            }
            else if(task.req){
                task.req.abort();
                task.req=null;
            }
        }
    }
    else{
        db.getAndRemove({id:id},function(err,data){
            if(!err){
                task=data;
                if(task && task.stat==="WAIT"){
                    self.waitSize--;
                }
                else if(task && task.stat==="PAUSE"){
                    self.pauseSize--;
                }
                else if(task && task.stat==="FAILED"){
                    self.failedSize--;
                }
                self.allSize--;
                self.emit("onAbort",[task]);
            }
        });
    }
};


//暂停
TaskManager.prototype.pause = function pause(id,callback) {
    var self = this;
    var task = self.runningMap[id];
    if (task) {
        global.logger.info("pause task:",task.bucket," ",task.key," ",task.type);
        if (task.stat === "RUNNING") {
            //上传
            if (task.type === "putObj") {
                if (task.upload) {
                    //分段任务
                    task.upload.pause(function () {

                        task.stat = "PAUSE";

                        if (task.upload.partSize) {
                            task.partSize = task.upload.partSize;
                        }
                        if (task.upload.queueSize) {
                            task.queueSize = task.upload.queueSize;
                        }
                        if (task.upload.completeMap) {
                            task.completeMap = task.upload.completeMap;
                        }
                        task.totalUploaded = 0;
                        if (task.upload.totalUploaded) {
                            task.totalUploaded = task.upload.totalUploaded;
                        }
                        //流引用
                        delete task["upload"];

                       self.finish(task, callback);
                    });

                } else if(task.req){
                    //不分段任务
                    task.stat = "PAUSE";
                    task.progress = "0%";
                    task.req.abort();
                    //流引用
                    delete task['req'];
                    self.finish(task, callback);
                }
            }
            //下载
            else if (task.type === "getObj") {
                if (task.download) {
                    task.download.pause(function (){
                        //关闭文件
                        task.stat="PAUSE";
                        if(task.download){
                            if(task.download.partSize){
                                task.partSize=task.download.partSize;
                            }
                            if(task.download.completeMap){
                                task.completeMap=task.download.completeMap;
                            }
                            if(task.download.totalDownloadedBytes){
                                task.totalDownloaded=task.download.totalDownloadedBytes;
                            }
                            delete task["download"];

                            self.finish(task,callback);
                        }
                    });
                }
                else if(task.req){
                    task.stat = "PAUSE";
                    task.req.removeAllListeners('httpDownloadProgress');
                    task.req.removeAllListeners('complete');
                    task.req.removeAllListeners('error');
                    task.req.abort();
                    delete task['req'];
                }
            }
        }
    }
    else {
        db.get({id:id},function(err,data){
            if(!err){
               task = data;
                if (task.stat === "WAIT") {
                    self.waitSize--;
                    self.pauseSize++;
                }
                else if (task.stat === "FAILED") {
                    self.failedSize--;
                    self.pauseSize++;
                }
                task.stat = "PAUSE";
                delete task['active'];
                db.updateStat(task,function(err,data){
                    if (callback) {
                        callback.call();
                    }
                });
            }
        });
    }
};

//开始
TaskManager.prototype.start = function start(id){
    var self=this;
    if(self.downloadPool){
        self.downloadPool.pauseAll=false;
    }
    for(var key in self.uploadBucketPools){
        self.uploadBucketPools[key].pauseAll=false;
    }
    db.get({id:id},function(err,data){
        if(!err)
            var task=data;
            if(task){
                if(task.stat==="PAUSE"||task.stat==="FAILED"){
                    if(task.stat==="PAUSE")
                    {
                        self.pauseSize--;
                    }
                    else if(task.stat==="FAILED")
                    {
                        self.failedSize--;
                    }
                    task.stat="WAIT";
                    self.waitSize++;
                    db.updateStat(task,function(err,data){
                        if(!err){
                            self.emit("onWait",[task]);
                        }
                    })
                }
            }
    });

};

//开始所有
TaskManager.prototype.startAll = function(callback){
    var self=this;
    self.setPauseAll(false);
    self.setAbortAll(false);

    var queryString = {userId : self.userId ,type:{"$ne":"recover"},"$or" : [ { "stat" : "PAUSE"} , { "stat" :"FAILED"}]};
    var param = {$set: {stat : "WAIT"}};
    db.updateProperties(queryString,param,function(err,data){
        if(!err){
        }
        if(callback)
            callback(err,data);
    });
};
//暂停所有
TaskManager.prototype.pauseAllTask = function(callback){
    var self=this;
    self.setPauseAll(true);
    global.logger.debug("pause all tasks!");
    var queryString = {userId : self.userId , "stat" : "WAIT"};
    var param = {$set: {stat : "PAUSE"}};
    db.updateProperties(queryString,param,function(err,data){
        if(callback)
            callback(err,data);
    });
};

//取消所有
TaskManager.prototype.abortAllTasks = function(callback){
    var self=this;
    var query = {userId : self.userId,type:"getObj"};
    global.logger.debug("Abort all tasks!");
    db.lists(query, function (err,data) {
        if(!err){
            for(var i in data){
                var filePath = path.normalize(data[i].path+data[i].id);
                if(fs.existsSync(filePath)){
                    try{
                        fs.unlink(filePath, function(err){});
                    }catch(err){
                        global.logger.error('unlink download file error',err);
                    }
                }
            }
        }else{
            global.logger.error('lists download tasks error',err);
        }
        var queryString = {userId : self.userId,stat:{$nin:['SUCCESS','RESTORING',"MD5SUCCESS"]}};
        db.delByQueryString(queryString,function(err,data){
            if(!err){
            }
            if(callback)
                callback(err,data);
        });
    });
};

TaskManager.prototype.setPauseAll = function(flag){
    var self=this;
    self.pauseAll=flag;
    if(self.downloadPool){
        self.downloadPool.pauseAll=flag;
    }
    for(var key in self.uploadBucketPools){
        self.uploadBucketPools[key].pauseAll=flag;
    }
};

TaskManager.prototype.setAbortAll = function(flag){
    var self=this;
    self.abortAll=flag;
    if(self.downloadPool){
        self.downloadPool.abortAll=flag;
    }
    for(var key in self.uploadBucketPools){
        self.uploadBucketPools[key].abortAll=flag;
    }
};





TaskManager.prototype.loadTaskNum = function loadTaskNum(callback){
    var self = this;
    var count = 0;
    db.count({userId:self.userId,stat:"WAIT"},function(err,data){
        if(!err)
            self.waitSize=data;
        count++;
        if(count===4){
            if(callback)callback();
        }
    });
    db.count({userId:self.userId,stat:"PAUSE"},function(err,data){
        if(!err)
            self.pauseSize=data;
        count++;
        if(count===4){
            if(callback)callback();
        }
    });
    db.count({userId:self.userId,stat:"FAILED"},function(err,data){
        if(!err)
            self.failedSize=data;
        count++;
        if(count===4){
            if(callback)callback();
        }
    });
    db.count({userId:self.userId},function(err,data){
        if(!err)
            self.allSize=data;
        count++;
        if(count===4){
            if(callback)callback();
        }
    });

    for(var key in self.uploadBucketPools){
        db.count({userId:self.userId,stat:"WAIT",bucket:key},function(err,data){
            if(!err){
                self.uploadBucketPools[key].waitSize = data;
            }
        });
    }
};

//创建或回收buckepool
TaskManager.prototype.manageBucketPools=function manageBucketPools(){
    var self=this;
    //根据等待队列中的上传和删除任务的桶创建桶任务池。
    var queryString={userId:self.userId,stat:"WAIT","$or" : [ { "type" : "create"} , { "type" :"putObj"} , { "type" :"delete"} , { "type" :"copy"} , { "type" :"move"}]};
    db.distinct("bucket",queryString,function(err,data){
        if(err){
            global.logger.error('DB distinct bucket is err:',err);
        }
        if(!err && data && data.length > 0 ) {
            var count = data.length;
            for(var i=0;i<count;i++){
                var bucketName = data[i];
                if(!self.uploadBucketPools[bucketName]){
                    var uploadBucketPool = new UploadBucketPool(bucketName,self.userId,self.s3,self.uploadConfig,db);
                    self.uploadBucketPools[bucketName] = uploadBucketPool;
                    self.uploadBucketPools[bucketName]
                        .on("progress", function(task, info){
                            self.emit("progress",[task, info]);
                        }).on("onStart", function (task) {
                            self.runningMap[task.id]=task;
                            self.emit("onStart",[task]);
                        }).
                        on("onWait", function (task) {
                            delete self.runningMap[task.id];
                            self.emit("onWait",[task]);
                        }).
                        on("onPause", function (task) {
                            delete self.runningMap[task.id];
                            self.emit("onPause",[task]);
                        }).
                        on("failed", function (task) {
                            delete self.runningMap[task.id];
                            self.emit("failed",[task]);
                        }).
                        on("onAbort", function (task) {
                            delete self.runningMap[task.id];
                            self.emit("onAbort",[task]);
                        }).
                        on("end", function (task) {
                            delete self.runningMap[task.id];
                            self.emit("end",[task]);
                        });
                }
                else{
                    self.uploadBucketPools[bucketName].lifecycle++;
                }
            }

        }
    });
    //回收无用的桶任务池（等待池中也无该桶任务且该bucketpool中无任务）
    for(var key in self.uploadBucketPools){
        if(self.uploadBucketPools[key].lifecycle>3 && self.uploadBucketPools[key].waitSize===0){
            for(var key in self.uploadBucketPools[key].runningMap){
                return;
            }
            delete self.uploadBucketPools[key];//删除桶任务池
        }
    }
};


//设置上传参数
TaskManager.prototype.setUploadConfig=function setUploadConfig(uploadConfig){
    var self=this;
    if(uploadConfig){
        if(uploadConfig.isMultiUpload){
            self.uploadConfig.isMultiUpload = uploadConfig.isMultiUpload;
        }else{
            self.uploadConfig.isMultiUpload = false;
        }
        if(uploadConfig.multiUploadNum){
            self.uploadConfig.multiUploadNum = uploadConfig.multiUploadNum;
        }
        if(uploadConfig.maxUploadTasks){
            global.maxUploadTasks = uploadConfig.maxUploadTasks;
        }
        else{
            global.maxUploadTasks=10;
            console.warn('maxUploadTasks is not undefined',uploadConfig);
        }
        if(uploadConfig.maxTps){
            self.uploadConfig.maxTps = uploadConfig.maxTps;
        }
        for (var key in self.uploadBucketPools) {
            if (uploadConfig.isMultiUpload) {
                self.uploadBucketPools[key].isMultiUpload = uploadConfig.isMultiUpload;
            }else{
                self.uploadBucketPools[key].isMultiUpload = false;
            }
            if (uploadConfig.multiUploadNum) {
                self.uploadBucketPools[key].multiUploadNum = uploadConfig.multiUploadNum;
            }
            if (uploadConfig.maxTps) {
                self.uploadBucketPools[key].maxTps = uploadConfig.maxTps;
            }
        }
    }
};

//设置下载参数
TaskManager.prototype.setDownloadConfig=function setDownloadConfig(downloadConfig){
    var self=this;
    if(downloadConfig){
        if(downloadConfig.isMultiDownload){
            self.downloadConfig.isMultiDownload = downloadConfig.isMultiDownload;
            self.downloadPool.isMultiDownload = downloadConfig.isMultiDownload;
        }else{
            self.downloadConfig.isMultiDownload = false;
        }

        if(downloadConfig.maxTaskNum){
            self.downloadConfig.maxTaskNum = downloadConfig.maxTaskNum;
            self.downloadPool.maxTaskNum = downloadConfig.maxTaskNum;
        }
    }
};


TaskManager.prototype.loadUploadRunningTaskNum = function loadUploadRunningTaskNum(){
    var self = this;
    var count = 0;
    for(var key in self.runningMap){
        if(self.runningMap[key].type != "getObj"){
            count++;
        }
    }
    global.uploadRunningSize = count;
};

//query任务数
TaskManager.prototype.countTasks = function (query,callback){
    var self=this;
    query.userId = self.userId;
    db.count(query,callback);
};

//query任务列表
TaskManager.prototype.listTasks = function (query,pageSize,pageNum,callback){
    var self=this;
    query.userId = self.userId;
    db.list(query,pageSize,pageNum,callback);
};
//等待任务列表
TaskManager.prototype.listWaitTasks = function (pageSize,pageNum,callback){
    var self=this;
    db.list({userId:self.userId,stat:"WAIT"},pageSize,pageNum,callback);
};
//暂停任务列表
TaskManager.prototype.listPauseTasks = function (pageSize,pageNum,callback){
    var self=this;
    db.list({userId:self.userId,stat:"PAUSE"},pageSize,pageNum,callback);

};
//失败任务列表
TaskManager.prototype.listFailedTasks = function (pageSize,pageNum,callback){
    var self=this;
    db.list({userId:self.userId,stat:"FAILED"},pageSize,pageNum,callback);
};
//所有任务列表
TaskManager.prototype.listAllTasks = function (pageSize,pageNum,callback){
    var self=this;
    db.list({userId:self.userId},pageSize,pageNum,callback);
};

//所有任务列表
TaskManager.prototype.listAllUserTasks = function (callback){
    db.lists({},callback);
};

//关闭DB
TaskManager.prototype.closeDB = function (callback){
    clearInterval(this.time);
    db.close(callback);
};


//清空DB
TaskManager.prototype.clearDB = function (callback){
    db.clear();
};

//模糊查询校验文件夹是否可删除
TaskManager.prototype.deletedCheckQuery = function(queryObj,callback){
    var self=this;
    queryObj.userId = self.userId;
    db.regexQuery(queryObj,function(err,data){
        if(!err){
        }
        if(callback)
            callback(err,data);
    });
};

//批量删除任务
TaskManager.prototype.deleteTasks = function (ids,callback){
    db.delSomeTasks(ids,callback);
};

//列举所有未完成的恢复任务
TaskManager.prototype.listAllRunRecoverTasks = function (stat,pageSize,pageNum,callback){
    var self=this;
    db.list({userId:self.userId,type:"recover",stat:stat},pageSize,pageNum,callback);
};

//批量更新任务
TaskManager.prototype.updatesRecoverTasks = function (ids,callback){
    var self = this;
    var queryString = {userId : self.userId,id: {"$in":ids}};
    var param = {$set: {stat : "SUCCESS",endTime:new Date()}};
    db.updateProperties(queryString,param,function(err,data){
        if(callback)
            callback(err,data);
    });
};

//模糊查询校验文件夹是否可删除
TaskManager.prototype.listAndGetResult = function(queryObj,result,callback){
    var self=this;
    queryObj.userId = self.userId;
    db.listAndGetResult(queryObj,result,function(err,data){
        if(!err){
        }
        if(callback)
            callback(err,data);
    });
};

AWS.util.mixin(TaskManager, AWS.SequentialExecutor);
module.exports = TaskManager;