var AWS = require('hws-sdk/lib/core');
var fs = require('fs');
var ChunkedReadStream = require('./ChunkedReadStream');
var obsError = require("./common/ObsError.js");
var CommonUtils = require("./common/CommonUtils.js");
var crypto = require('crypto');
var async = require('async');
function BucketPool() {

}

constructor: function BucketPool(bucketName, userId, s3, uploadConfig, db) {
    var self = this;
    AWS.SequentialExecutor.call(self);
    this.bucket = bucketName;
    this.userId = userId;
    this.s3 = s3;
    this.db = db;

    if (uploadConfig) {
        this.isMultiUpload = uploadConfig.isMultiUpload;
        this.multiUploadNum = uploadConfig.multiUploadNum;
        this.maxTps = 15;
    }

    this.pauseAll = false;
    this.abortAll = false;
    this.runningMap = {}; //运行集合
    this.waitSize = 0; //等待任务数
    self.lifecycle = 0; //本桶池的生命周期。回收时判断至少3个周期的才被回收
};

//读取并执行任务
BucketPool.prototype.run = function run() {
    var self = this;
    if (!global.maxUploadTasks) {
        global.maxUploadTasks = 10;
    }
    if (global.maxUploadTasks > global.uploadRunningSize && this.waitSize > 0 && !this.pauseAll) {
        self.waitSize--;
        global.uploadRunningSize++;

        //查询指定用户指定桶，类型为上传和删除的等待任务
        var queryString = {
            bucket: self.bucket,
            userId: self.userId,
            stat: "WAIT",
            "$or": [{
                "type": "create"
            }, {
                "type": "putObj"
            }, {
                "type": "delete"
            }, {
                "type": "copy"
            }, {
                "type": "move"
            }]
        };
        self.db.getAndModify(queryString, function(err, taskRecord) {
            if (!err && taskRecord) {
                var task = taskRecord.value;
                if (task) {
                    task.startTime = new Date();
                    global.logger.debug("execute task:", task.bucket, " ", task.key, " ", task.type);
                    task.stat = "RUNNING";
                    self.runningMap[task.id] = task;
                    self.emit("onStart", [task]);
                    if (task.type === "create") {
                        var params = {};
                        if (global.serverEncryption) {
                            params = {
                                Bucket: task.bucket,
                                Key: task.key,
                                ServerSideEncryption: 'aws:kms'
                            };
                        } else {
                            params = {
                                Bucket: task.bucket,
                                Key: task.key
                            };
                        }
                        self.s3.putObject(params, function(err, data) {
                            if (err) {
                                global.logger.error('[s3] put obj--->', err);
                                task.stat = "FAILED";
                                task.errMessage = err.code;
                            }

                            self.finish(task);
                        });
                    } else if (task.type === "putObj") {
                        self.upload(task);
                    } else if (task.type === "delete") {
                        var params = {
                            Bucket: task.bucket,
                            Key: task.key
                        };
                        self.s3.deleteObject(params, function(err) {
                            if (err) {
                                global.logger.error('[s3] del obj', err);
                                task.stat = "FAILED";
                                task.errMessage = err.code;
                            }
                            self.finish(task);
                        });
                    } else if (task.type === "copy" || task.type === "move") {
                        self.copyMoveObj(task);
                    }
                } else {
                    self.waitSize++;
                    global.uploadRunningSize--;
                }
            } else {
                self.waitSize++;
                global.uploadRunningSize--;
            }
        });
    } else {}
};

/**
 * 拷贝移动具体执行逻辑
 * @param task
 */
BucketPool.prototype.copyMoveObj = function(task) {
    var uploadBucketPool = this;
    var headParams = {
        Bucket: task.sourceBucket,
        Key: task.sourceKey
    };
    uploadBucketPool.s3.headObject(headParams, function(err, data) {
        if (err) {
            global.logger.error("[s3] copy obj", err, "params:", headParams);
            task.stat = "FAILED";
            if (err.statusCode === 404) {
                err.code = "NoSuchKey";
            }
            task.errMessage = CommonUtils.format(global.langData.obs_task_pasteObjectFail_msg, obsError(err.code).description);
            uploadBucketPool.finish(task);
            return;
        }

        var source = "/" + task.sourceBucket + "/" + task.sourceKey;
        var params = {
            Bucket: task.bucket,
            Key: task.key,
            CopySource: encodeURIComponent(source)
        }
        if (data && data.SSEKMSKeyId && data.ServerSideEncryption) {
            params.SSEKMSKeyId = data.SSEKMSKeyId;
            params.ServerSideEncryption = data.ServerSideEncryption;
        }

        uploadBucketPool.s3.copyObject(params, function(err, data) {
            if (err) {
                global.logger.error("[s3] copy obj", err, "params:", params);
                task.stat = "FAILED";
                task.errMessage = CommonUtils.format(global.langData.obs_task_pasteObjectFail_msg, obsError(err.code).description);
                uploadBucketPool.finish(task);
                return;
            }
            //todo 处理ACL问题
            var getAclParams = {
                Bucket: task.sourceBucket,
                Key: task.sourceKey
            }
            uploadBucketPool.s3.getObjectAcl(getAclParams, function(err, data) {
                if (!err && data) {
                    data.Owner.ID = global.curOnwerId;
                    data.Owner.DisplayName = global.curDisplayName;
                    var putAclParams = {
                        Bucket: task.bucket,
                        Key: task.key,
                        AccessControlPolicy: {
                            Grants: data.Grants,
                            Owner: data.Owner
                        }
                    };
                    uploadBucketPool.s3.putObjectAcl(putAclParams, function(err, data) {
                        if (err) {
                            global.logger.error("[s3] copy obj，put objACL", err, "params:", putAclParams);
                            task.stat = "FAILED";
                            var totalMsg = CommonUtils.format(global.langData.obs_task_pasteObjectSuccess_setAclFail_msg, obsError(err.code).description)
                            task.errMessage = totalMsg;
                            uploadBucketPool.finish(task);
                            return;
                        }
                        //todo 处理move类型的delete操作
                        if (task.type === "move") {
                            var headParams = {
                                Bucket: task.bucket,
                                Key: task.key
                            };
                            uploadBucketPool.s3.headObject(headParams, function(err, data) {
                                if (!err) {
                                    //todo 删除source对象
                                    var delObjParams = {
                                        Bucket: task.sourceBucket,
                                        Key: task.sourceKey
                                    };
                                    uploadBucketPool.s3.deleteObject(delObjParams, function(err) {
                                        if (err) {
                                            global.logger.error("[s3] move delObj", err);
                                            task.stat = "FAILED";
                                            task.errMessage = CommonUtils.format(global.langData.obs_task_pasteObjectSuccess_delSourceobjectFail_msg, obsError(err.code).description);
                                        }
                                        uploadBucketPool.finish(task);
                                    });
                                } else {
                                    global.logger.error("[s3] move headObj", err);
                                    task.stat = "FAILED";
                                    task.errMessage = CommonUtils.format(global.langData.obs_task_pasteObjectSuccess_getTargetObjectInfoFail_msg, obsError(err.code).description);
                                    uploadBucketPool.finish(task);
                                }
                            });
                        } else {
                            uploadBucketPool.finish(task);
                        }
                    })
                } else {
                    global.logger.error("[s3] copy obj，get objACL", err, "params:", getAclParams);
                    task.stat = "FAILED";
                    task.errMessage = CommonUtils.format(global.langData.obs_task_pasteObjectSuccess_getSourceObjectAclFail_msg, obsError(err.code).description);
                    uploadBucketPool.finish(task);
                }
            })
        });
    })
}

//上传文件
BucketPool.prototype.upload = function upload(task) {
    var self = this;
    if (fs.existsSync(task.path)) {
        //如果对象KEY大于1023个字节，则直接让任务失败
        var objectNameChar = new Buffer(task.key);
        var objectNameLength = objectNameChar.length;
        if (objectNameLength > 1023) {
            global.logger.error('[local] put obj error code:KeyTooLongError,message:The absolute path of an object cannot exceed 1023 bytes.');
            task.stat = "FAILED";
            task.errMessage = "KeyTooLongError";
            self.finish(task);
            return false;
        }

        var stat = fs.lstatSync(task.path);
        // 上传文件配置ACL权限
        var ACL = 'bucket-owner-full-control';
        var sseKmsKeyId = 'arn:aws:kms:'+global.curRegion+':'+global.curOnwerId+':key/'+global.serverEncryptionKeyId;
        if (stat.size === 0) {
            //文件大小为0，文件夹也走这里
            var params = {};
            if (global.serverEncryption) {
                params = {
                    Bucket: task.bucket,
                    Key: task.key,
                    ServerSideEncryption: 'aws:kms',
                    StorageClass: task.storagePolicy,
                    ACL: ACL
                };
            } else {
                params = {
                    Bucket: task.bucket,
                    Key: task.key,
                    StorageClass: task.storagePolicy,
                    ACL: ACL
                };
            }
            this.s3.putObject(params, function(err, data) {
                if (err) {
                    global.logger.error('[s3] put obj failed-->', task, err);
                    task.stat = "FAILED";
                    task.errMessage = err.code;
                } else {}
                self.finish(task);
            });
        } else if (stat.size <= 65536) {
            //文件大小为size<=64K
            try {
                var body = fs.readFileSync(task.path);
                var params = {};
                if (global.serverEncryption) {
                    params = {
                        Bucket: task.bucket,
                        Key: task.key,
                        Body: body,
                        ServerSideEncryption: 'aws:kms',
                        StorageClass: task.storagePolicy,
                        ACL: ACL
                    };
                } else {
                    params = {
                        Bucket: task.bucket,
                        Key: task.key,
                        Body: body,
                        StorageClass: task.storagePolicy,
                        ACL: ACL
                    };
                }
                var req;
                req = this.s3.putObject(params).
                on('httpUploadProgress', function(evt) {
                    if (task.stat === "RUNNING")
                        self.emit("progress", [task, evt]);
                });
                if (req) {
                    task.req = req;
                    global.logger.info('[mult uploader] <64 --upload send key param:', task.key, params);
                    req.send(function(err, data) {
                        if (err) {
                            global.logger.error('[s3]--put obj <64kb-->key stat err:',task.key,task.stat, err);
                            //networkerror可能导致回调2次，一次是根因，还有一次是根因导致的本端socket关闭。如果第2次异常出现时任务已经重新开始执行，则不能再次重试。
                            if (task.req !== req)
                            {
                                global.logger.warn("[putobj] networkerrorS. and this err must be socket hang up.")
                                return;
                            }
                            delete task["req"];
                            task.stat = "FAILED";
                            task.errMessage = err.code;

                            //用户取消任务
                            if (err.code === "RequestAbortedError") {
                                task.stat = "ABORT";
                            }
                        }
                        self.finish(task);
                    });
                }
            } catch (err) {
                if (err) {
                    global.logger.error('[upload pool]--catch-->', err);
                    task.stat = "FAILED";
                    task.errMessage = err.code;
                }
                self.finish(task);
            }
        } else {
            //是否分片上传
            if (self.isMultiUpload) {
                var params = {
                    Bucket: task.bucket,
                    Key: task.key,
                    Body: task.path,
                    StorageClass: task.storagePolicy,
                    ACL: ACL
                };
                if (global.serverEncryption) {
                    params.ServerSideEncryption = 'aws:kms';
                }
                var partSize = Number(self.multiUploadNum);
                var options = {
                    partSize: partSize * 1024 * 1024,
                    queueSize: 4,
                    leavePartsOnError: true,
                    completeMap: {},
                    totalUploaded: 0
                };

                //有任务号
                if (task.UploadId) {
                    //如果已经采用了服务端加密，则不改变
                    if (task.serverEncryption) {
                        params.ServerSideEncryption = task.serverEncryption;
                    }
                    options.UploadId = task.UploadId;
                    options.completeMap = task.completeMap;
                    options.totalUploaded = task.totalUploaded;
                    options.completeMap1 = task.completeMap1;
                }

                if (task.partSize) {
                    options.partSize = task.partSize;
                }
                if (task.queueSize) {
                    options.queueSize = task.queueSize;
                }
                global.logger.info('[mult uploader] >64 --upload send>>', task.key, params, options);
                var upload = this.s3.upload(params, options, function(err, data) {
                    if (err) {
                        global.logger.error('[s3]--put obj >64kb-->key stat err:',task.key,task.stat, err);
                        //networkerror可能导致回调2次，一次是根因，还有一次是根因导致的本端socket关闭。
                        // 如果第2次异常出现时任务已经重新开始执行，则不能再次重试
                        if (task.upload !== upload)
                        {
                            global.logger.warn("[putobj] networkerrorS. and this err must be socket hang up.",err)
                            return;
                        }
                        delete task["upload"];
                        task.stat = "FAILED";
                        task.errMessage = err.code;
                        //用户取消任务
                        if (err.code === "RequestAbortedError") {
                            task.stat = "ABORT";
                        } else if (err.code === 'NoSuchUpload') {
                            global.logger.warn('[NoSuchUpload]', task.UploadId, err);
                            delete task.UploadId; //删除uploadId导致已经上传的段无效了。
                            task.stat = "WAIT";
                        }
                    }
                    self.finish(task);
                });
                upload.on('successCreateMultipartUpload', function(UploadId) {
                    task.UploadId = UploadId;
                    task.serverEncryption = params.ServerSideEncryption;
                }).
                on('httpUploadProgress', function(evt) {
                    self.emit("progress", [task, evt]);
                }).on('partComplete', function(part) {
                    task.completeMap = this.completeMap;
                    task.totalUploaded = this.totalUploaded;
                    var uTask={};
                    for(var k in this.task){
                        if(k != 'upload'){
                            uTask[k]=this.task[k];
                        }
                    }
                    self.db.updateStat(uTask,function(err){
                        if(err){
                            global.logger.error(err);
                        }
                    });
                });
                upload.task = task;
                if(upload){
                    task.upload = upload;
                }
            } else {
                //默认为多段上传，目前这个分支无效
                try {
                    if (global.serverEncryption && global.isChunk) {
                        var stat = fs.statSync(task.path);
                        var rs = fs.createChunkedReadStream(task.path, {
                            highWaterMark: 65536
                        }).on("open", function(fd) {
                            var len = rs.calculateStreamContentLength(stat.size);
                            var params = {
                                Bucket: task.bucket,
                                Key: task.key,
                                Body: rs,
                                ServerSideEncryption: 'aws:kms',
                                DecodedContentLength: stat.size + '',
                                ContentLength: len + '',
                                ContentSHA256: 'STREAMING-AWS4-HMAC-SHA256-PAYLOAD',
                                StorageClass: 'REDUCED_REDUNDANCY',
                                ContentEncoding: 'aws-chunked'
                            };
                            var req;
                            req = self.s3.putObject(params).
                            on('httpUploadProgress', function(evt) {
                                if (task.stat === "RUNNING")
                                    self.emit("progress", [task, evt]);
                            });
                            if (req) {
                                task.req = req;
                                req.send(function(err, data) {
                                    if (task.stat === "RUNNING") {
                                        if (err) {
                                            global.logger.error('[s3]--put obj-->RUNNING-->', err);
                                            delete task["req"];
                                            task.stat = "FAILED";
                                            task.errMessage = err.code;
                                            if (task.errMessage == "AccessDenied") {
                                                task.progress = "0%";
                                            }

                                            //用户取消任务
                                            if (err.code === "RequestAbortedError") {
                                                task.stat = "ABORT";
                                            }
                                        } else {}
                                        self.finish(task);
                                    }
                                });
                            }
                        }).on("error", function(err) {
                            global.logger.error('[fs]--stream-onerror-->', err);
                            task.stat = "FAILED";
                            task.errMessage = err.code;
                            self.finish(task);
                        });
                    } else {
                        var rs = fs.createReadStream(task.path).on("open", function(fd) {
                            var params = {};
                            if (global.serverEncryption) {
                                params = {
                                    Bucket: task.bucket,
                                    Key: task.key,
                                    Body: rs,
                                    ServerSideEncryption: 'aws:kms'
                                };
                            } else {
                                params = {
                                    Bucket: task.bucket,
                                    Key: task.key,
                                    Body: rs
                                };
                            }
                            var req;
                            req = self.s3.putObject(params).
                            on('httpUploadProgress', function(evt) {
                                if (task.stat === "RUNNING")
                                    self.emit("progress", [task, evt]);
                            });
                            if (req) {
                                task.req = req;
                                req.send(function(err, data) {
                                    if (task.stat === "RUNNING") {
                                        if (err) {
                                            global.logger.error('[s3]--put obj-->RUNNING-->', err);
                                            delete task["req"];
                                            task.stat = "FAILED";
                                            task.errMessage = err.code;
                                            if (task.errMessage == "AccessDenied") {
                                                task.progress = "0%";
                                            }

                                            //用户取消任务
                                            if (err.code === "RequestAbortedError") {
                                                task.stat = "ABORT";
                                            }
                                        } else {}
                                        self.finish(task);
                                    }
                                });
                            }
                        }).on("error", function(err) {
                            global.logger.error('[fs]--stream-onerror-->', err);
                            task.stat = "FAILED";
                            task.errMessage = err.code;
                            self.finish(task);
                        });
                    }
                } catch (err) {
                    if (err) {
                        global.logger.error('[fs]--stream-onerror-->', err);
                        task.stat = "FAILED";
                        task.errMessage = err.code;

                    }
                    self.finish(task);
                }
            }
        }
    } else {
        //文件不存在
        task.stat = "FAILED";
        task.errMessage = "NoSuchFile";
        global.logger.error('[fs]--NoSuchFile-->', task);
        self.finish(task);
    }
};


//任务结束，包含正常结束、暂停、失败、取消
BucketPool.prototype.finish = function finish(task, callback) {
    var self = this;
    global.logger.info("uploadbucketPool task finish.....key,stat,errMessage:",task.key, task.stat,task.errMessage)
    if (this.runningMap[task.id]) {
        global.uploadRunningSize = (global.uploadRunningSize-- > 0) ? global.uploadRunningSize : 0; //global.uploadRunningSize最小值为0
        delete this.runningMap[task.id];
        var times = 0;
        if (task.startTime) {
            times = new Date().getTime() - task.startTime.getTime();
        }
        if (task.stat === "RUNNING") {
            if (task.type === "putObj" && global.isMD5check === 'Y') {
                var stat = fs.lstatSync(task.path);
                var isDirectory = stat.isDirectory();
                if(isDirectory){
                    task.stat = 'SUCCESS';
                    delete task['active'];
                    delete task['speed'];
                    delete task['spTime'];
                    delete task['upload'];
                    delete task['completeMap'];
                    delete task['req'];
                    task.endTime = new Date();
                    self.db.updateStat(task, function(err, data) {
                        if (!err) {
                            self.emit("end", [task]);
                        } else {
                            global.logger.error('[DB]--error', err);
                        }
                    });
                }else{
                    task.stat = 'MD5Check';
                    var md5num = crypto.createHash('md5');
                    var size = 0 ;//保存文件大小
                    var chunk =fs.createReadStream(task.path).on('data',function(chunks){
                        md5num.update(chunks);
                        size+=chunks.length;
                    }).on('end',function () {
                        var headParams = {
                            Bucket: task.bucket,
                            Key:task.key
                        }
                        s3.headObject(headParams,function (err,data) {
                            if(data.ContentLength == size){
                                var  HeaderMD5 = md5num.digest('hex');
                                chunk.destroy();
                                var params = {//复制对象，将对象MD5设置为自定义数据
                                    Bucket : task.bucket,
                                    Key:  task.key,
                                    MetadataDirective : "REPLACE",
                                    CopySource:  task.bucket + "/" + encodeURIComponent(task.key),
                                    HeaderMD5:HeaderMD5,
                                    ACL:'bucket-owner-full-control'
                                };
                                s3.copyObject(params,function (err,data) {
                                    task.stat = "MD5SUCCESS";
                                    delete task['active'];
                                    delete task['speed'];
                                    delete task['spTime'];
                                    delete task['upload'];
                                    delete task['completeMap'];
                                    delete task['req'];
                                    task.endTime = new Date();
                                    self.db.updateStat(task, function(err, data) {
                                        if (!err) {
                                            self.emit("end", [task]);
                                        } else {
                                            global.logger.error('[DB]--error', err);
                                        }
                                    });
                                })
                            }else{
                                task.stat = "FAILED";
                                delete task['active'];
                                delete task['speed'];
                                delete task['spTime'];
                            }
                        })
                    })
                    global.successLogger.info("upload", 'key:', task.key, "size:", task.size, "accountName:", global.currentUserName, "duration:", times, "status:", "MD5Success");
                }
            } else{
                task.stat = 'SUCCESS';
                delete task['active'];
                delete task['speed'];
                delete task['spTime'];
                delete task['upload'];
                delete task['completeMap'];
                delete task['req'];
                task.endTime = new Date();
                self.db.updateStat(task, function(err, data) {
                    if (!err) {
                        self.emit("end", [task]);
                    } else {
                        global.logger.error('[DB]--error', err);
                    }
                });
            }
        } else if (task.stat === "PAUSE") {
            delete task['active'];
            delete task['speed'];
            delete task['spTime'];
            self.db.updateStat(task, function(err, data) {
                if (!err) {
                    self.emit("onPause", [task]);
                } else {
                    global.logger.error('[DB]--error', err);
                }
            });

        } else if (task.stat === "FAILED") {
            delete task['speed'];
            delete task['spTime'];
            if (task.type === "putObj") {
                global.successLogger.info("upload", 'key:', task.key, "size:", task.size, "accountName:", global.currentUserName, "duration:", times, "errCode:", task.errMessage, "status:", "Failed");
            } else if (task.type === "delete") {
                global.successLogger.info("delete", 'key:', task.key, "size:", task.size, "accountName:", global.currentUserName, "duration:", times, "errCode:", task.errMessage, "status:", "Failed");
            }
            if (task.errMessage === 'NetworkingError') {
                task.stat = 'WAIT';
                task.errMessage = '';
                self.db.updateStat(task, function(err, data) {
                    if (!err) {
                        self.emit("onWait", [task]);
                    } else {
                        global.logger.error('[DB]--error', err);
                    }
                });
            } else {
                task.endTime = new Date();
                self.db.updateStat(task, function(err, data) {
                    if (!err) {
                        self.emit("failed", [task]);
                    } else {
                        global.logger.error('[DB]--error', err);
                    }
                });
            }
        } else if (task.stat === "WAIT") {
            self.db.updateStat(task, function(err, data) {
                if (!err) {

                } else {
                    global.logger.error('[DB]--error', err);
                }
            });

        } else if (task.stat === "ABORT") {
            self.db.delById(task.id, function(err, data) {
                if (!err) {
                    self.emit("onAbort", [task]);
                } else {
                    global.logger.error('[DB]--error', err);
                }
            });
        }
    }
    if (callback) {
        callback.call();
    }

};

BucketPool.prototype.loadWaitTaskNum = function loadWaitTaskNum(callback) {
    var self = this;
    self.db.count({
        bucket: self.bucket,
        userId: self.userId,
        stat: "WAIT",
        "$or": [{
            "type": "create"
        }, {
            "type": "putObj"
        }, {
            "type": "delete"
        }, {
            "type": "copy"
        }, {
            "type": "move"
        }]
    }, function(err, data) {
        if (!err)
            self.waitSize = data;
    });
};


AWS.util.mixin(BucketPool, AWS.SequentialExecutor);
module.exports = BucketPool;