/**
 * Created by yk on 2017/1/19.
 */
var async = require('async');
exports.notifySplashScreen = function () {
    var net = require('net');

    var client = new net.Socket();
    try {
        client.connect(1337, '127.0.0.1', function () {
            console.log('Connected');
            client.write("loaded");
        });
        client.on('data', function (data) {
            console.log('Received: ' + data);
        });
        client.on('error', function () {
            global.logger.error("Can not find splash screen. Connect failed");
            return false;
        });
        client.on('close', function () {
            console.log('Connection closed');
        });
    } catch (err) {
        global.logger.error("Connect splash screen program failed.", err);
    }
}

/**
 * 处理上传下载进度异常显示问题
 * @param newPer 新计算出的进度
 * lastPer 原进度
 */
exports.handlePer = function (newPer, lastPer) {
    //如果原进度不存在，则直接返回新进度
    if (!lastPer) {
        return newPer;
    }
    lastPer = lastPer.replace(/%/g, "");
    //因某些异常场景会导致进度大于100%,出现此种情况时将进度强制改为99.99
    if (newPer >= 100) {
        newPer = 99.99;
        return newPer;
    }

    //避免进度回退
    if (lastPer > newPer) {
        newPer = lastPer;
    }
    return newPer;
};

/**
 * 兼容新老出包方案，DT已经把package.json提取到上一层目录了
 * @param module
 */
exports.perspectiveRequirePackageCfg = function (module) {
    var rqrModule = null;
    try {
        rqrModule = require(module);
    } catch (err) {
        //DT应该查找require('../package.json')，其他局点是的require('./package.json')
        //当打包过程全部nonzip化以后，则全部用require('../package.json')
        global.logger.warn(module, "can not find module...try to find it from top dir.")
        rqrModule = require("../../." + module);
    }
    return rqrModule;
}

exports.isValidJson = function (str) {

}

/**
 * 合并词条，以targetLangArr为基线，如果enWordsArr中有多的key，则把enWordsArr[key]添加到targetLangArr中
 * var b = {"a":"bba","bb":"bbbbb","c":"bbc","d":"dddddd"}
 * var a = {"a":"a","b":"bb","c":"ccc"}
 * result = {a: "a", b: "bb", c: "ccc", bb: "bbbbb", d: "dddddd"}
 * @param enWordsArr 默认的英语词条
 * @param targetLangArr 对应语言的词条
 * @returns {*}
 */
exports.mergeDefaultLang = function (enWordsArr, targetLangArr) {
    var result = targetLangArr;
    for (var key in enWordsArr) {
        result[key] = targetLangArr[key] ? targetLangArr[key] : enWordsArr[key];
    }
    return result;
}

/**
 * 校验Policy的值是否合法
 * @param policyData
 * @returns {boolean}
 */
exports.checkPolicy = function (policyData) {
    var val = policyData.trim();
    if ("" !== val) {
        try {
            var data = eval('(' + val + ')');
        }
        catch (e) {
            return false;
        }
        if ("{}" == val.split("\n").join("").split(" ").join("")) {
            return false;
        }
        return true;
    } else {
        return true;
    }
}

/**
 * 获取不同局点的差异性词条
 * @param UCDVersion 局点信息
 * @param UCDVersion 词条key
 * @returns {boolean}
 */
exports.getDiffVersionLang = function (UCDVersion, key) {
    return key + "_" + UCDVersion;
}

/**
 * 根据温冷存储银蛇
 * @param key
 * @returns {*}
 */
exports.translateStorageClass = function (key) {
    var viewToIntfMap = {
        STANDARD: "Standard",
        STANDARD_IA: "Nearline",
        GLACIER: "Coldline",
        Standard: "STANDARD",
        Nearline: "STANDARD_IA",
        Coldline: "GLACIER"
    }
    return viewToIntfMap[key];
}

exports.termStorageClass = function (key) {
    var viewToIntfMap = {};
    if (global.UCDVersion === "HEC") {
        viewToIntfMap = {
            Standard: global.langData.obs_term_standard_lable,
            Nearline: global.langData.obs_term_nearline_lable_HEC,
            Coldline: global.langData.obs_term_coldline_lable_HEC,
            STANDARD: global.langData.obs_term_standard_lable,
            STANDARD_IA: global.langData.obs_term_nearline_lable_HEC,
            GLACIER: global.langData.obs_term_coldline_lable_HEC
        }
    } else {
        viewToIntfMap = {
            Standard: global.langData.obs_term_standard_lable,
            Nearline: global.langData.obs_term_nearline_lable,
            Coldline: global.langData.obs_term_coldline_lable,
            STANDARD: global.langData.obs_term_standard_lable,
            STANDARD_IA: global.langData.obs_term_nearline_lable,
            GLACIER: global.langData.obs_term_coldline_lable
        }
    }
    return viewToIntfMap[key];
}
/**
 * 获取任务类型
 * @param key
 */
exports.convertTaskType = function (key) {
    var taskType = {
        allType: global.langData.obs_term_allType_list,
        getObj: global.langData.down,
        putObj: global.langData.upload,
        "delete": global.langData.delete,
        recover: global.langData.obs_term_restore_button,
        SUCCESS: global.langData.obs_term_success_label,
        FAILED: global.langData.failed,
        RUNNING: global.langData.running,
        WAIT: global.langData.wait,
        PAUSE: global.langData.paused,
        RESTORING: global.langData.obs_term_restoring_lable,
        copy: global.langData.obs_term_copyObject_button,
        move: global.langData.obs_term_moveObject_button,
        MD5Check:global.langData.obs_term_md5Checking,
        MD5SUCCESS:global.langData.obs_term_success_label
    }
    return taskType[key];
}

/**
 * 将数据库数据写入到csv文件中
 * @param writeStream 可写流
 * @param data 数据
 * @param cb
 */
exports.writeTaskData2CSV = function (writeStream, data, cb) {
    async.eachLimit(data, 20, function (item, callback) {
        var dbData = new Buffer(item.key + "," + item.type + "," + item.bucket + ',' + item.size + "," + item.startTime + "," + item.endTime + ",\n", 'utf8');
        writeStream.write(dbData, function (err) {
            callback(err);
        });
    }, function (err) {
        if (err) {
            global.logger.error('Write dbdata to csv failed:', err);
        }
        if (cb) {
            cb();
        }
    });
};

exports.compareExpiresDays = function (value, expires) {
    var expires = new Date(expires);
    var now = new Date();
    var date3 = expires.getTime() - now.getTime(); //时间差的毫秒数 //计算出相差天数
    var days = Math.floor(date3 / (24 * 3600 * 1000));
    if (days > value) {
        return false;
    } else {
        return true;
    }
};

exports.format = function () {
    if (arguments) {
        var str = arguments[0];
        for (var i = 0; i < arguments.length; i++) {
            var reg = new RegExp("\\{" + i + "\\}", "gm");
            str = str.replace(reg, arguments[i + 1]);
        }
        return str;
    }
    return "";
};

exports.substr32 = function (value) {
    if (typeof(value) === "string") {
        var len = 32;
        if (value.length > len) {
            return value.substring(0, len) + "...";
        } else {
            return value;
        }
    } else {
        return value;
    }
};

exports.encodeForHTML = function (input) {
    var div = window.document.createElement('div');
    div.innerText = input;
    return div.innerHTML;
};