/*
本地文件遍历
 */

var util = require("util"),EventEmitter =require("events").EventEmitter;
var fs = require("fs");

function LocalPathIterator(option,subDirs){
    this.subDirs= new Array();
    if(subDirs){
        for(var i=0;i<subDirs.length ;i++){
            this.subDirs.push(subDirs[i]);
        }
    }

    this.type ="local";
    this.option = option;
    this.path="";
    this.files = new Array();
    this.paused=false;
    this.foundSize=option.foundSize || 0;
    this.busy =false;
    this.waiting=false;
    this.finished=false;
    this.lastValidateCount=Date.now();
}

//添加事件监听机制
util.inherits(LocalPathIterator,EventEmitter);
LocalPathIterator.prototype.pause = function(){
    this.paused=true;

};
LocalPathIterator.prototype.setWaiting = function(v){
    global.logger.info("setWaiting", v);
    this.waiting =v;
};
LocalPathIterator.prototype.resume = function(){
    this.paused=false;
};

LocalPathIterator.prototype.save = function(path,cb){
    var self = this;
    //解决暂停时可能遗漏文件的问题
    //需要暂停任务完成，处于空闲状态时保存状态
    if(self.busy){
        setTimeout(function(){
            self.save.apply(self,[path,cb]);
        },50);
        return;
    }
    var data={
        option:self.option,
        dirs:self.subDirs,
        foundSize:self.foundSize || 0,
        paused:self.paused
    };
    if(!fs.existsSync(path)){
        fs.mkdirSync(path);
    }
    if(!fs.existsSync(path+"iterators/")){
        fs.mkdirSync(path+"iterators/");
    }

    fs.writeFile(path+"iterators/"+this.option.id+".itr",JSON.stringify(data),function(err){
        if(cb){
            cb(err);
        }
    });

};

LocalPathIterator.prototype.load= function(itrFile,cb){
    var self = this;
    fs.readFile(itrFile,function(err,data){
        if(err){
              console.error('read file error',err);
            if(cb){
                cb(err,self);
                return;
            }
        }
        else{
            try{
                var jsonData= JSON.parse(data);
                self.paused= false;
                self.waiting=false;
                self.busy=false;
                self.foundSize = jsonData.foundSize || 0;
                self.option=jsonData.option;
                self.subDirs = jsonData.dirs;
                setTimeout(function(){
                    fs.unlinkSync(itrFile);
                },2000);
            }catch(e){
                if(cb){
                    cb(e,self);
                    return;
                }
            }


        }
        if(cb){
            cb(err,self);
        }
    });
};

LocalPathIterator.prototype.start = function(){
    var self = this;
    if(this.paused){
        return false;
    }
    if(this.waiting || this.busy){
        setTimeout(function(){
            self.start.apply(self);
        },2000);
        self.emit("waiting",this);
        return;
    }

    if(this.subDirs.length === 0){
        this.emit("finished",this);
        this.finished=true;
        return;
    }
    this.path=this.subDirs.shift();
    self.busy =true;
    fs.readdir(this.path,function(err,dirs){
        if(self.path[self.path.length-1] === '/'){
            self.path=self.path.substring(0,self.path.length-1);
        }
        
        //创建文件夹
        var floderInfo = {path:self.path+"/"};
        self.emit("floderFound",self,floderInfo);

        if(err){
            console.error("readdir error",err);
            self.emit("warning",err,self.path);
            self.start();
            self.busy =false;
            return;
        }
        for(var i = 0;i<dirs.length;i++){
            var dir = dirs[i];
            try{
                var stat = fs.lstatSync(self.path+"/"+dir);

                if(stat.isDirectory()){
                    if(stat.isSymbolicLink()){
                        global.logger.warn('ingone dir ',self.path+"/"+dir);
                    }
                    else{
                        self.subDirs.push(self.path+"/"+dir);
                    }
                }else if(stat.isSymbolicLink()){
                    global.logger.warn('ingone dir or file',self.path+"/"+dir);
                }
                else{
                    var fileInfo = {path:self.path+"/"+dir,size:stat.size};
                    self.files.push(fileInfo);

                    if(!self.foundSize){
                        self.foundSize=0;
                    }
                    self.foundSize ++;
                    self.emit("fileFound",self,fileInfo);
                }
            }
            catch (e){
                global.logger.error('Iterator path catch err:',self.path,e);
            }

        }
        self.busy =false;
        setTimeout(function(){
            self.start.apply(self);
        },1);

    });
};

LocalPathIterator.prototype.next = function(){
    if(this.files.length === 0){
        return false;
    }
    return this.files.shift();
};


LocalPathIterator.prototype.addPath = function(path){
    this.subDirs.push(path);
};

LocalPathIterator.prototype.cancelAll = function(){
};


exports = module.exports = LocalPathIterator;

