/**
    包括桶和对象所有操作接口以及账号配置和系统配置
 */

var util = require("util"),EventEmitter =require("events").EventEmitter;
//添加事件监听机制
util.inherits(BucketAndObjectAction,EventEmitter);
var IndexAction = require("./indexAction");
var indexAction = new IndexAction();
var CommonUtils = require("../common/CommonUtils.js")
var rubbishDays = 0;
var self;
var logger;
function BucketAndObjectAction(s3,rubbishDays){
    this.s3 = s3;
    this.rubbishDays = rubbishDays;
    self = this;
}

util.inherits(BucketAndObjectAction,EventEmitter);

/*
   获取桶列表
 */
BucketAndObjectAction.prototype.listBucket= function(callBack){
    this.s3.listBuckets(function (err, data) {
        if (callBack) {
            callBack(err, data);
        }
    });
}

/*
 获取桶内对象
 bucketName：桶名
 isDelimiter：
 prefix：前缀
 marker：取数据的首位置
 */
BucketAndObjectAction.prototype.listObject= function(bucketName, isDelimiter, prefix, success, marker ){
    var self = this;
    var params = {Bucket: bucketName, Prefix: prefix, Marker: marker};
    if (isDelimiter === "true") {
        params.Delimiter = "/";
    }
    //不是根目录并且marker不存在(是第一页)
    if((prefix.length > 0) && (!marker || marker === '')){
        var headParams = {Bucket: bucketName,Key:prefix};
        var maxNumber = 0;
        //判断当前列举的文件夹对象本身是否存在，如果存在，则列举条数加1
        self.s3.headObject(headParams, function (err, data) {
            if(err) {
                global.logger.error("list Object head Object err: ", err, "headPrarams: ", headParams);
            }
            //列举外部桶内对象会出现Forbidden错误
            if(err && err.code !== 'NotFound' && err.code !== 'BadRequest' && err.code !== 'Forbidden'){
                if(success){
                    success(err);
                }
                return false;
            }
            if(data) {
                maxNumber = 1;
            }
            var maxKeys = parseInt(global.objPerPageSize) + maxNumber;
            params.MaxKeys = maxKeys;
            global.logger.info('listObject params:',params);
            self.s3.listObjects(params, function (err, data) {
                if(success){
                    success(err, data);
                }
            });
        });
    }else{
        params.MaxKeys = global.objPerPageSize;
        global.logger.info('listObject params:',params);
        self.s3.listObjects(params, function (err, data) {
            if(success){
                success(err, data);
            }
        });
    }
};

BucketAndObjectAction.prototype.deleteBucket = function(bucketName, cb) {
    var start = new Date()
    global.logger.debug('deleteBucket Name:',bucketName);
    this.s3.deleteBucket({Bucket: bucketName}, function (err, data) {
        global.logger.debug('deleteBucket cost:', (new Date()) - start);
        if(cb){
            cb(err, data);
        }
    });
}


/*
    桶名称校验
 */
BucketAndObjectAction.prototype.bucketNameCheck = function bucketNameCheck(bucketName,languageJson) {
    var regex = new RegExp("^[a-z0-9]{1}[a-z0-9-.]{1,61}[a-z0-9]{1}$");
    var regIp = /^(25[0-5]|2[0-4]\d|[01]?\d\d?)\.(25[0-5]|2[0-4]\d|[01]?\d\d?)\.(25[0-5]|2[0-4]\d|[01]?\d\d?)\.(25[0-5]|2[0-4]\d|[01]?\d\d?)$/i;
    var regBlank = /^\s*$/g;
    var str = "";
    if (regBlank.test(bucketName)) {
        str = languageJson.bucketNameNotNull;
        return str;
    }
    if (!regex.test(bucketName)) {
        str = languageJson.bucketNameCriterion;
        return str;
    }
    if (regIp.test(bucketName)) {
        str = languageJson.bucketNameNotIP;
        return str;
    }
    if (bucketName.indexOf("..") > -1 || bucketName.indexOf(".-") > -1 || bucketName.indexOf("-.") > -1) {
        str =languageJson.bucketNameNotSpecail;
    }
    return str;
};


BucketAndObjectAction.prototype.createBucket = function createBucket(storageType, bucketName, region, finished) {
    var params = {
        Bucket: bucketName,
        StorageType: CommonUtils.translateStorageClass(storageType)
    };
    if(region){
        params.CreateBucketConfiguration = {
            LocationConstraint: region
        }
    }
    global.logger.info('createBucket params:',params);
    s3.createBucket(params, function (err, data) {
        if (finished) {
            finished(err, data);
        }
    });
};

/**
 * 对象恢复接口
 * @param bucket 桶名称
 * @param key 对象key
 * @param days 恢复多少天，int类型
 * @param tier 恢复速度，可选类型：Standard | Bulk | Expedited
 * @param callback 回调方法，接口返回值
 */
BucketAndObjectAction.prototype.restoreObject = function (bucket, key, days, tier, callback) {
    var params = {
        Bucket: bucket,
        Key: key,
        RestoreRequest: {
            Days: days, /* required */
            GlacierJobParameters: {
                Tier: tier /* required */
            }
        }
    }
    global.logger.info("restore object, params: ", params);
    s3.restoreObject(params, function(err, data){
        if(err) {
            global.logger.error("restore object failed...", err);
            if(callback) {
                callback(err);
            }
        }else{
            if(callback) {
                callback(null, data);
            }
        }
    })
}
/*
创建文件夹名称规则校验
 */
BucketAndObjectAction.prototype.folderNameCharacter = function (str,languageJson,cb,path){
    if(!str||str.trim().length===0)
    {
        if(cb){
            cb(languageJson.InputFolderName,str);
        }
        return false;
    }
    var str = str.trim();
    var parrent = /[<>?:*\\|\"]/g;
    if (parrent.test(str)) {
        str = str.replace(parrent,"");
        if(cb){
            cb(languageJson.inputVerification,str);
        }
    }else if(str.indexOf('.')===0||str.substring(str.length-1,str.length)==='.'||str.indexOf('/')===0||str.substring(str.length-1,str.length)==='/') {
        str = str.replace(/\./g, "");
        if (cb) {
            cb(languageJson.describe_object_ruleThree, str);
        }
    }else if(str.indexOf("//") != -1){
        if(cb){
            cb(languageJson.inputVerification3,str);
        }
    }else{
        var folderNameChar=new Buffer(path+str.trim());
        var folderNameLength=folderNameChar.length;
        if(folderNameLength>1023)
        {
            if(cb){
                cb(languageJson.folderNameTooLarge,str,"folderNameTooLarge");
            }
            return false;
        }else{
            if(cb){
                cb('',str);
            }
        }
    }
};


//新建子目录
BucketAndObjectAction.prototype.createFolder = function (bucketName, key, callback) {
    var params = {Bucket: bucketName, Key: key};
    global.logger.info('createFolder params:',params);
    s3.putObject(params, function (err, data) {
        if(callback){
            callback(err, data);
        }
    });
};

//修改碎片过期时间
BucketAndObjectAction.prototype.updateRubbishDays = function (rubbishDays) {
    rubbishDays = rubbishDays;
};

//列举桶内碎片列表
BucketAndObjectAction.prototype.listMultipartUploads = function (bucketName, prefix, marker , uploadIdMarker,pageNum,cb,parameter) {
    if(!pageNum){
        pageNum = 1;
    }
    if(parameter == undefined) {
        return;
    }
    var arr = parameter.arr;
    var params = {Bucket: bucketName, Prefix: prefix, Delimiter: "", KeyMarker: marker, MaxUploads: 20,UploadIdMarker:uploadIdMarker};
    global.logger.info('listMultipartUploads params:',params);
    s3.listMultipartUploads(params,function(err,data){
        if(err){
            if(cb){
                cb(err,parameter);
            }
        }else{
            for (var index in data.Uploads) {
                var obj = data.Uploads[index];
                if(compareTwoTime(gmtDate2LocaleDate(obj.Initiated),parameter.rubbishDays)){
                    if(arr.length<pageNum*20){
                       arr.push(obj);
                    }else{
                        parameter.pageMultipartContext.NextKeyMarker = obj.Key;
                        parameter.pageMultipartContext.NextUploadIdMarker = obj.UploadId;
                        parameter.pageMultipartContext.refresh = true;
                        break;
                    }
                }
            }
            if(data.IsTruncated&&arr.length<pageNum*20){
                parameter.arr = arr;
                self.listMultipartUploads(bucketName, prefix, data.NextKeyMarker , data.NextUploadIdMarker,pageNum,cb,parameter);
            }else{
                if(data.IsTruncated&&!parameter.pageMultipartContext.refresh){
                    parameter.pageMultipartContext.NextKeyMarker = data.NextKeyMarker;
                    parameter.pageMultipartContext.NextUploadIdMarker = data.NextUploadIdMarker;
                    parameter.pageMultipartContext.refresh = true;
                }
                parameter.pageMultipartContext.nowPage = parameter.pageMultipartContext.nowPage+1;
                if(cb){
                    cb(err,parameter);
                }
            }
        }
    });
};

//列举桶内每个碎片文件数据
BucketAndObjectAction.prototype.listParts = function (bucketName,id,key,partNumberMarker,size,count,cb){
    var params = {
        Bucket: bucketName,
        UploadId : id,
        Key:key,
        MaxParts:100,
        PartNumberMarker:partNumberMarker
    };
    s3.listParts(params , function (err,data) {
        if(err){
            global.logger.error(err);
            if(cb){
                cb(err,"");
            }
        }else{
            if(data.IsTruncated){
                for(var index in data.Parts){
                    size += data.Parts[index].Size;
                    count++;
                }
                self.listParts(bucketName,id,key,data.NextPartNumberMarker,size,count,cb);
            }else{
                for(var index in data.Parts){
                    size += data.Parts[index].Size;
                    count++;
                }
                var data = {};
                data.size = size;
                data.count = count;
                if(cb){
                    cb("",data);
                }
            }
        }

    });
};

//删除单个碎片
BucketAndObjectAction.prototype.abortMultipartUpload = function (bucketName,id,key,isRefresh,cb){
    var params = {
        Bucket: bucketName,
        UploadId : id,
        Key:key
    };
    global.logger.info('abortMultipartUpload params:',params);
    s3.abortMultipartUpload(params , function (err,data) {
        if(err){
            if(isRefresh&&err.code==='NoSuchUpload'){
                if(cb){
                    cb('');
                }
            }else{
                if(cb){
                    cb(err,data);
                }
            }
        }else{
            if(cb){
                cb(err,data);
            }
        }
    });
};

//删除所有碎片
BucketAndObjectAction.prototype.deleteAllMultipartUploads = function (bucketName, prefix, marker, uploadId,cb,isShowDelRubErrMsg,rubbishDays) {
    var params;
    params = {Bucket: bucketName, Prefix: prefix, Delimiter: "", KeyMarker: marker, MaxUploads: 10,UploadIdMarker:uploadId};
    global.logger.info('deleteAllMultipartUploads params:',params);
    s3.listMultipartUploads(params,function(err,data){
        if(err){
            if(cb){
                cb(err,false);
            }
        }else{
            var n = 0;
            var m = 0;
            if(data.IsTruncated){
                for (var index in data.Uploads) {
                    var obj = data.Uploads[index];
                    if(compareTwoTime(gmtDate2LocaleDate(obj.Initiated),rubbishDays)){
                        m++;
                       var start = new Date().getTime();
                        self.abortMultipartUpload(bucketName,obj.UploadId,obj.Key,true,function(err){
                            var end = new Date().getTime();
                            global.logger.debug('time ==== ',end-start);
                            if(err&&isShowDelRubErrMsg){
                                isShowDelRubErrMsg = false;
                                if(cb){
                                    cb(err,false);
                                }
                            }else{
                                n++;
                                if(m === n){
                                    global.logger.debug('asdfadf',bucketName,prefix,data.NextKeyMarker,data.NextUploadIdMarker,'',rubbishDays);
                                    self.deleteAllMultipartUploads(bucketName,prefix,data.NextKeyMarker,data.NextUploadIdMarker,cb,'',rubbishDays);
                                }
                            }
                        });
                    }
                }
            }else{
                if(data.Uploads.length>0){
                    for (var index in data.Uploads) {
                        var obj = data.Uploads[index];
                        if(compareTwoTime(gmtDate2LocaleDate(obj.Initiated),rubbishDays)){
                            m++;
                            var obj = data.Uploads[index];
                            var start = new Date().getTime();
                            self.abortMultipartUpload(bucketName,obj.UploadId,obj.Key,true,function(err){
                                var end = new Date().getTime();
                                global.logger.debug('time1 ==== ',end-start);
                                if(err&&isShowDelRubErrMsg){
                                    isShowDelRubErrMsg = false;
                                    if(cb){
                                        cb(err,false);
                                    }
                                }else{
                                    n++;
                                    if(m === n){
                                        if(cb){
                                            cb(err,true);
                                        }
                                    }
                                }
                            });
                        }
                    }
                }else{
                    if(cb){
                        cb('',true);
                    }
                }
            }
        }
    });
};

/*
 外部桶校验
 */
BucketAndObjectAction.prototype.extBucketNameCheck = function (bucketName,languageJson,callback) {
    var tem = true;
    this.listBucket(function (err, data) {
        if(!err)
        {
            for(var index in data.Buckets){
                var obj = data.Buckets[index];
                if(obj.Name === bucketName){
                    tem = false;
                    break;
                }
            }
            if(callback){
                callback('',tem);
            }
        }else if(err.code === "AccessDenied"){
            callback(err)
        }else{
            callback('languageJson.InternalError');
        }
    });
};

/*
 删除外部桶
 */
BucketAndObjectAction.prototype.deleteExtBucket = function (bName,ExtBucket,currentUserId,callback) {
    global.logger.info('deleteExtBucket params:',bName,ExtBucket,currentUserId);
    var arr = [];
    var obj = {};
    for(var i in ExtBucket){
        if(ExtBucket[i].id===currentUserId){
            obj = ExtBucket[i];
            arr = ExtBucket[i].bucketArr;
            ExtBucket.splice(i,1);
            break;
        }
    }
    for(var index in arr){
        if(arr[index]===bName){
            arr.splice(index,1);
            break;
        }
    }
    if(arr.length!==0){
        obj.bucketArr = arr;
        ExtBucket.push(obj);
    }
    if(callback){
        callback(ExtBucket);
    }
};

/*
 获取桶的Website
 */
BucketAndObjectAction.prototype.getBucketWebsite = function (bName,callback) {
    var params = {Bucket:bName};
    s3.getBucketWebsite(params, function (err, data) {
        if(callback){
            callback(err,data);
        }
    });
};


/**
 * 用户选择已有对象进行服务端加密--复制对象
 */
BucketAndObjectAction.prototype.copyObject = function (bucketName,storagePolicy,objectName,KMS,sseKMSKeyId,callback)
{
    var params = {
        Bucket : bucketName,
        Key:  objectName,
        MetadataDirective : "COPY",
        CopySource:  bucketName + "/" + encodeURIComponent(objectName),
        StorageClass:storagePolicy
    };
    if(KMS){
        params.ServerSideEncryption = 'aws:kms';
        params.SSEKMSKeyId = sseKMSKeyId;
    }

    global.logger.info('copy object params:',params);
    s3.copyObject(params, function(err, data)
    {
        if (callback)
        {
            callback(err, data);
        }

    });
};

//判断对象是否存在
BucketAndObjectAction.prototype.headObject = function (bucketName,key,cb){
    var params = {
        Bucket: bucketName,
        Key:key
    };
    global.logger.info('headObject params:',params);
    s3.headObject(params , function (err,data) {
        if(err){
            global.logger.error(err);
            if(err.code == "NotFound") {
                err.code = "NoSuchKey"
            }
            if(cb){
                cb(err,data,key);
            }
        }else{
            if(cb){
                cb(err,data,key);
            }
        }
    });
};
//判断桶是否存在
BucketAndObjectAction.prototype.headObjectWithMD5 = function (bucketName,key,unique,cb){
    var params = {
        Bucket: bucketName,
        Key:key
    };
    s3.headObject(params , function (err,data) {
        if(err){
            global.logger.error(err);
            if(err.code == "NotFound") {
                err.code = "NoSuchKey"
            }
            if(cb){
                cb(err,data,key,unique);
            }
        }else{
            if(cb){
                cb(err,data,key,unique);
            }
        }
    });
};

/*
 比较两个时间大小
 */
function compareTwoTime (times,rubbishDays){
    var date1=new Date(times);
    var date2 = new Date();
    date2 = new Date(date2.getTime()-rubbishDays*24*60*60*1000);
    return date1<date2;
};

function gmtDate2LocaleDate (dateStr) {
    var creationDate=dateStr;
    var date=new Date(dateStr);

    var year=date.getFullYear();
    var m=date.getMonth()+1;
    var mm=m<10?"0"+m:m;
    var d=date.getDate();
    var dd=d<10?"0"+d:d;
    var h=date.getHours();
    var hh=h<10?"0"+h:h;
    var min=date.getMinutes();
    var minm=min<10?"0"+min:min;
    var s=date.getSeconds();
    var ss=s<10?"0"+s:s;

    var createTime=year+"-"+mm+"-"+dd+" "+hh+":"+minm+":"+ss+" ";


    var d=new Date(dateStr);
    var dGMT=-(d.getTimezoneOffset());
    var hourGMT=0;
    if((dGMT/60)<0){
        if(dGMT/60>10){
            hourGMT=-(-Math.ceil(dGMT/60)+1);
        }else{
            hourGMT=-(-Math.ceil(dGMT/60));
        }

    }else{
        hourGMT=Math.floor(dGMT/60);
    }
    hourGMT=hourGMT<0?""+hourGMT:"+"+hourGMT;

    var hG1=hourGMT.substr(0,1);
    var hG2=hourGMT.substr(1,hourGMT.length);
    var hG3=hG2<10?"0"+hG2:hG2;
    var hG=hG1+hG3;
    var minGMT=dGMT%60;
    if(dGMT%60<0){
        minGMT=-(dGMT%60);
    }
    var mG=minGMT<10?"0"+minGMT:minGMT;
    createTime=" "+createTime+"GMT"+hG+":"+mG;
    return createTime;
}
module.exports = BucketAndObjectAction;