/*
 任务控制接口
 */
var util = require("util"),EventEmitter =require("events").EventEmitter;
var async = require('async');
var os = require('os');
var fs = require('fs');
var LocalPathIterator = require('../common/LocalPathIterator.js');
var intervalTime = 10;//定时器间隔时间
//添加事件监听机制
util.inherits(TaskControlAction,EventEmitter);

var rootPath=process.execPath;
rootPath = rootPath.replace(/\\/g, "/");
rootPath = rootPath.substring(0,rootPath.lastIndexOf('/')+1);

var userPath = false;
var queryTypeDelOrPut={ "type" :"putObj"};
var queryTypeGet={ "type" : "getObj"};
var Const = require('../common/constants.js');
 var commonUtils = require('../common/CommonUtils.js');
function TaskControlAction(s3,taskManager){
    this.winType = os.platform();
    this.s3 = s3;
    this.taskManager = taskManager;
    this.fileQueue=[];
    this.iterators = [];
    var self= this;
    this.iteratorsObject = true;

    this.lastValidateCount=Date.now();
}

util.inherits(TaskControlAction,EventEmitter);

/*
 uuid生存算法
 */
function guid() {
    var S4 = function () {
        return Math.floor(
            Math.random() * 0X10000 /*65536*/
        ).toString(16)
    };
    return S4() + S4() + "-" + S4() + "-" + S4() + "-" + S4() + "-" + S4() + S4() + S4();
}

TaskControlAction.prototype.closeInterval=function(){
    clearInterval(this.time);
    clearInterval(this.taskManager.time);
    clearInterval(this.taskManager.time1);
    clearInterval(this.taskManager.loadWaitTasksItv);
};

TaskControlAction.prototype.resumeIterators=function(){
     var self = this;
    userPath = rootPath+this.taskManager.userId+'/';
    if(!fs.existsSync(userPath)){
        fs.mkdirSync(userPath);
    }
    if(!fs.existsSync(userPath+"iterators/")){
        fs.mkdirSync(userPath+"iterators/");
    }
    fs.readdir(userPath+'iterators/',function(err,files) {
        if(err || !files){
            console.warn(err,files);
            return;
        }

        for (var i = 0; i < files.length; i++) {
            if(!files[i]){
                console.warn('ign file',i);
                continue;
            }
            var pathIterator = new LocalPathIterator({});
            pathIterator.lastValidateCount=Date.now();
            pathIterator.queue=self.fileQueue;
            pathIterator.on('fileFound', function(itr){
                var realPath = itr.option.path.replace(/:/g, "$");
                //上传对象push到任务队列
                var file = itr.next();

                var key = itr.option.prefix + file.path.substring(realPath.length);
                var fInfo={
                    bucket:itr.option.bucket,
                    key:key,
                    size:file.size,
                    path:file.path
                };
                self.uploadFile(fInfo.bucket, fInfo.key, fInfo.path);
                if(itr.lastValidateCount+1000<Date.now()){
                    self.taskManager.countTasks(queryTypeDelOrPut,function(err,ret){

                       itr.setWaiting(ret>2000);
                    });
                    itr.lastValidateCount=Date.now();
                }
            });
            pathIterator.on('floderFound',function(itr){
                var realPath = itr.option.path.replace(/:/g, "$");
                //上传对象push到任务队列
                var floder = itr.path;

                var key = itr.option.prefix + floder.substring(realPath.length)+"/";
                var fInfo={
                    bucket:itr.option.bucket,
                    key:key,
                    size:0,
                    path:floder
                };
                self.uploadFile(fInfo.bucket, fInfo.key, fInfo.path);
                if(itr.lastValidateCount+1000<Date.now()){
                    self.taskManager.countTasks(queryTypeDelOrPut,function(err,ret){
                        itr.setWaiting(ret > 2000);
                    });
                    itr.lastValidateCount=Date.now();
                }
            });
            pathIterator.on('finished',function(){
                for(var k in self.iterators){
                    if(this.option.id == self.iterators[k].option.id){
                        self.iterators[k]=null;
                        delete  self.iterators[k];
                        break;
                    }
                }
            });
            pathIterator.on('waiting',function(itr){
                if(itr.lastValidateCount+3*1000<Date.now()){
                    self.taskManager.countTasks(queryTypeDelOrPut,function(err,ret){

                        itr.setWaiting(ret>2000);
                    });
                    itr.lastValidateCount=Date.now();
                }
            });
            pathIterator.load(userPath + 'iterators/' + files[i], function (err, itr) {
                if(err){
                    console.error('load error ',err);
                    return ;
                }
                itr.resume();

                itr.start();
                self.iterators.push(pathIterator);
            });
        }

    });
};


//删除一个对象
TaskControlAction.prototype.deleteObject = function (bucket, key ,size) {
    var self = this;
    if(!bucket || !key){
        return;
    }
    var uuid = guid();
    var task =
    {
        id: uuid,
        type: "delete",
        bucket: bucket,
        key: key,
        size: size,
        stat:"WAIT"
    };
    this.taskManager.push(task);
}
/*
 * 批量删除(文件夹)
 * bucket 桶名
 * parentPath 路径以"/"结尾
 * prefix 路径以"/"结尾
 */
TaskControlAction.prototype.deletePath = function (bucket, prefix,marker,num,cb) {
    var self = this;
    if(!self.iteratorsObject){
        return false;
    }
    if(!num){
        num = 0;
    }
    var params = {Bucket: bucket, Prefix: prefix, Delimiter: "", Marker: marker, MaxKeys: 100};
    if(!this.taskManager.abortAll){
        self.s3.listObjects(params, function (err, data) {
            if (err) {
                if(num<4){
                    num++;
                    setTimeout(function () {
                        self.deletePath(bucket, prefix, marker,num,cb);
                        global.logger.error('this is deletePath function error = ',num);
                    },num*60000);
                }else{
                    if(cb){
                        cb(err);
                    }
                }
                global.logger.error(err);
            }
            else{
                num = 0;
                if(data.Contents.length === 0){
                    self.deleteObject(bucket, prefix ,0);
                    return false;
                }
                if (data.IsTruncated){
                    async.mapSeries(data.Contents, function(object, callback) {
                        setTimeout(function() {
                            var key = object.Key;
                            self.deleteObject(bucket, key ,object.Size);
                            if(self.taskManager.abortAll)
                                callback("abort by user");
                            else
                                callback();
                        }, 5);
                    }, function(err,results) {
                        if(err){
                            global.logger.error(err);
                        }
                    });
                    self.deletePath(bucket, prefix, data.NextMarker,num,cb);
                }else{
                    async.mapSeries(data.Contents, function(object, callback) {
                        setTimeout(function() {
                            var key = object.Key;
                            self.deleteObject(bucket, key ,object.Size);
                            if(self.taskManager.abortAll)
                                callback("abort by user");
                            else
                                callback();
                        }, 5);
                    }, function(err,results) {
                        if(err){
                            global.logger.error(err);
                        }
                    });
                }
            }
        });
    }
};

/*
 上传文件
 */
TaskControlAction.prototype.uploadFile = function(bucket, key, filePath, storageType) {
    var self = this;
    if(!this.taskManager.abortAll) {

        filePath = filePath.replace(/[']/g, '\'');
        key = key.replace(/:/g, "$");

        fs.lstat(filePath, function (err, data) {
            if (err) {
                global.logger.error(err);
                var errCode = err.code;
                var errNo = err.errno;
                errNo = errNo.substring(1, errNo.length);
                errCode = errCode + errNo;
                if (bucket && key && filePath) {
                    var uuid = guid();
                    var task =
                    {
                        id: uuid,
                        type: "putObj",
                        bucket: bucket,
                        size: 0,
                        progress: "0%",
                        stat: "FAILED",
                        key: key,
                        path: filePath,
                        errMessage: errCode,
                        storagePolicy: storageType
                    };
                    self.taskManager.push(task);
                }
            } else {
                if (bucket && key && filePath) {
                    var size = data.size;
                    var uuid = guid();
                    var task =
                        {
                            id: uuid,
                            type: "putObj",
                            bucket: bucket,
                            size: size,
                            progress: "0%",
                            stat: "WAIT",
                            key: key,
                            path: filePath,
                            storagePolicy: storageType
                        };
                    if(self.pauseAll){
                        task.stat='PAUSE';
                    }
                    self.taskManager.push(task);
                }
            }
        });
    }
};

/*
 上传文件夹
 */
TaskControlAction.prototype.uploadPath = function(bucket ,prefix ,path, storageType) {

    global.logger.info('[upload Path]',bucket ,prefix ,path, storageType);

    var start = path.substring(0, path.lastIndexOf("/", path.length - 2) + 1);
    var now = new Date();
    var self = this;
    var pathIterator = new LocalPathIterator({id:'itr-' + now.getTime(), bucket:bucket, prefix:prefix, path:start}, [path]);

    pathIterator.on('fileFound', function(itr){
        var realPath = itr.option.path.replace(/:/g, "$");
        //上传对象push到任务队列
        var file = itr.next();

        var key = itr.option.prefix + file.path.substring(realPath.length);
        var fInfo = {
            bucket:itr.option.bucket,
            key:key,
            size:file.size,
            path:file.path,
            storageType:storageType
        };
        self.uploadFile(fInfo.bucket, fInfo.key, fInfo.path, fInfo.storageType);
    });

    pathIterator.on('floderFound',function(itr){
        var realPath = itr.option.path.replace(/:/g, "$");
        //上传对象push到任务队列
        var floder = itr.path;

        var key = itr.option.prefix + floder.substring(realPath.length) + "/";
        var fInfo = {
            bucket:itr.option.bucket,
            key:key,
            size:0,
            path:floder,
            storageType:storageType
        };
        self.uploadFile(fInfo.bucket, fInfo.key, fInfo.path, fInfo.storageType);
    });

    pathIterator.on('finished',function(){
        global.logger.info("Iterator run finished.")
        for(var k in self.iterators){
            if(this.option.id == self.iterators[k].option.id){
                self.iterators[k]=null;
                delete  self.iterators[k];
                break;
            }
        }
    });

    pathIterator.start();

    self.iterators.push(pathIterator);
};
/*
 检查下载文件存放路径是否存在
 */
TaskControlAction.prototype.pathCheck = function(paths, callback) {
    var fs = require('fs');
    var path = require('path');
    paths = path.normalize(paths);
    fs.exists(paths, function (exist) {
        if (callback) {
            callback(exist);
        }
    })
};

/*
 下载文件
 bucket:bucketname
 key:key like 'a/b/c.txt'
 filepath:文件全路径like'd:/c.txt'
 */
TaskControlAction.prototype.downloadFile = function(bucket,key,filePath,size,etag) {
    var self = this;
    if (bucket && key && filePath) {
        var uuid = guid();
        var task =
        {
            id: uuid,
            type: "getObj",
            bucket: bucket,
            key: key,
            path: filePath,
            size: size,
            progress: "0%",
            stat: "WAIT",
            etag:etag
        };
        this.taskManager.push(task);
    }
};

/*
 下载文件夹
 */
TaskControlAction.prototype.downloadPath = function(bucket, parentPath,prefix, marker, path,num,cb) {
    var self = this;
    if(!self.iteratorsObject){
        return false;
    }
    if(!num){
        num = 0;
    }
    if(parentPath.charAt(parentPath.length-1)!='/'){
        parentPath += "/";
    }
    if(path.charAt(path.length-1)!='/'){
        path += "/";
    }

    if(!this.taskManager.abortAll){
        var params = {Bucket: bucket, Prefix: prefix, Marker: marker, MaxKeys: 100};
        s3.listObjects(params, function (err, data) {
            if (err) {
                //超时错误重试
                if(err.hasOwnProperty('code') && err.code == 'NetworkingError'){
                    if(num<4){
                        num++;
                        setTimeout(function () {
                            self.downloadPath(bucket, parentPath, prefix, marker, path,num,cb);
                            global.logger.error('this is downloadPath function error = ',num);
                        },num*60000);
                    }else{
                        if(cb){
                            cb(err);
                        }
                    }
                }else{
                    //不是超时的错误直接返回
                    cb(err);
                }
                global.logger.error(err);
            }
            else {
                if (data.IsTruncated){
                    num = 0;
                    async.mapSeries(data.Contents, function(object, callback) {
                        setTimeout(function() {
                            var key = object.Key;
                            if (key[key.length - 1] != '/') {
                                var subKey = key.substring(parentPath.length-1, key.length);
                                self.downloadFile(bucket, key, path + subKey, object.Size);
                            }else{
                                var subKey = key.substring(parentPath.length-1, key.length);
                                self.downloadFile(bucket, key, path + subKey, 0);
                            }
                            if(self.taskManager.abortAll)
                                callback("abort by user");
                            else
                                callback();
                        }, intervalTime);
                    }, function(err,results) {
                        if(err){
                            global.logger.error(err);
                        }
                    });
                    self.downloadPath(bucket, parentPath, prefix, data.NextMarker, path,num,cb);
                }
                else{
                    /**
                     * 顺序执行，一个完了才执行下一个。
                     */
                    async.mapSeries(data.Contents, function(object, callback) {
                        setTimeout(function() {
                            var key = object.Key;
                            if (key[key.length - 1] != '/') {
                                var subKey = key.substring(parentPath.length-1, key.length);
                                self.downloadFile(bucket, key, path + subKey, object.Size);
                            }else{
                                var subKey = key.substring(parentPath.length-1, key.length);
                                self.downloadFile(bucket, key, path + subKey, 0);
                            }
                            if(self.taskManager.abortAll)
                                callback("abort by user");
                            else
                                callback();
                        }, intervalTime);
                    }, function(err,results) {
                        if(err){
                            global.logger.error(err);
                        }
                    });
                }
            }
        });
    }
};


//开始任务
TaskControlAction.prototype.startTask = function (id) {
    this.taskManager.start(id);
};
//开始所有任务
TaskControlAction.prototype.startAllTasks = function (callback) {
    this.taskManager.startAll(callback);
    this.resumeIterators();
};

//暂停任务
TaskControlAction.prototype.pauseTask = function (id) {
    this.taskManager.pause(id);
};

//暂停所有任务
TaskControlAction.prototype.pauseAllTasks = function (callback) {
    var self = this;
    if(self.taskManager){
        userPath = rootPath+this.taskManager.userId+'/';
        async.each(this.iterators,function(itr,cb){
            if(itr){
                itr.pause();
                itr.save(userPath,function(){
                    cb();
                });
            }else{
                cb();
            }
        },function(err){
            for(var i = 0;i<self.iterators.length;i++){
                self.iterators[i]=null;
                delete self.iterators[i];
            }
            //暂停任务
            self.taskManager.setPauseAll(true);
            //暂停运行中的任务
            var tasks = self.taskManager.runningMap;
            for (var id in tasks) {
                self.taskManager.pause(id);
            }

            self.taskManager.pauseAllTask(callback);

        });
    }else{
        if(callback){
            callback();
        }
    }
};

//取消任务
TaskControlAction.prototype.abortTask = function (id, callback) {
    this.taskManager.abort(id);
    this.taskManager.on("onAbort",function(data){
        if(callback){
            callback(data)
        }
    });
};

//取消所有任务
TaskControlAction.prototype.abortAllTasks = function (callback) {
    this.taskManager.setAbortAll(true);
    var self = this;
    for(var i = 0;i<self.iterators.length;i++){
        if(self.iterators[i]){
            self.iterators[i].pause();
        }
        self.iterators[i]=null;
        delete self.iterators[i];
    }
    userPath = rootPath+this.taskManager.userId+'/';
    fs.readdir(userPath+'iterators/',function(err,files) {
        if(err || !files){
            console.warn(err,files);
            return;
        }
        for (var i = 0; i < files.length; i++) {
            if(!files[i]){
                console.warn('ign file',i);
                continue;
            }
            fs.unlink(userPath + 'iterators/' + files[i],function(err){
                console.warn('unlink',err);
            });
        }

    });
    self.fileQueue = [];
    var tasks = self.taskManager.runningMap;
    for (var id in tasks) {
        self.taskManager.abort(id);
    }
    //取消等待、暂停和失败任务
    self.taskManager.abortAllTasks(callback);

};

//切换用户时停止递归遍历UDS对象方法
TaskControlAction.prototype.updateIteratorsObject = function (stat) {
    this.iteratorsObject = stat;
};

/**
 * 检查数据库数据是否需要清理，如果超过50W则清理部分数据
 * @param num 每次查询条数
 * @param total 总共清理多少条记录
 */
TaskControlAction.prototype.checkDBData = function (){
    var self = this;
    self.taskManager.countTasks({stat:"SUCCESS"}, function (err,data) {
        global.logger.warn("Current completed task num: " + data);
        if(data && data > Const("NUM_MONGODB_TASKS")){
            global.logger.warn("Completed task num=" + data + ", over than maxinum（"+Const("NUM_MONGODB_TASKS")+"), start to execute clearDBData()");
            self.clearDBData();
        }
    });
};

/**
 * 清理数据库数据
 * @param num 每次查询条数
 * @param total 总共清理多少条记录
 */
TaskControlAction.prototype.clearDBData = function (){
    var self = this;
    var isStill = true;//是否继续递归
    var exePath=process.execPath;//文件路径
    exePath = exePath.replace(/\\/g, "/");
    exePath = exePath.substring(0,exePath.lastIndexOf('/')+1);
    var date = new Date().toISOString().replace(/:/g,"").replace(/-/g,"");
    date = date.substring(0,date.lastIndexOf('.'));

    //每次都新生成一个csv文件
    var writeStream = fs.createWriteStream(exePath+'history/historyDBData'+date+'.csv',{encoding:'utf8'});
    var headData = new Buffer("Name,Type,BucketName,Size,StartTime,endTime,\n",'utf8');
    writeStream.write(new Buffer('\xEF\xBB\xBF','binary'));
    writeStream.write(headData);
    var delNum = 0;//已删除任务个数

    //递归执行清理操作，每次从数据库取num条数据
    async.whilst(
        function() {
            return isStill && (delNum < Const("NUM_CLEAR_NUM"));
        },
        function(cb) {
            var query={"stat":"SUCCESS"};
            self.taskManager.listTasks(query,Const("NUM_MONGODB_LIMITS"),1,function(err,data){
                if(data && data.length > 0){
                    delNum += data.length;
                    //将任务备份到csv文件中
                    commonUtils.writeTaskData2CSV(writeStream,data, function () {
                        var ids = [];
                        for(var x in data){
                            ids.push(data[x].id);
                        }
                        var startTime = new Date().getTime();
                        self.taskManager.deleteTasks(ids, function () {
                            console.log('times:',new Date().getTime() - startTime);
                            cb();
                        });
                    });
                }else{
                    isStill = false;
                    cb();
                }
            });
        },
        function(err) {
            // 3s have passed
            global.logger.info('Clear dbdata success!');
            writeStream.end();
        }
    );
};

/*
  添加恢复任务
 */
TaskControlAction.prototype.recoverObject = function(bucket, key, size,stat,errCode,days,tier) {
    var self = this;
    var uuid = guid();
    var task =
    {
        id: uuid,
        type: "recover",
        bucket: bucket,
        size: size || 0 ,
        progress: "0%",
        stat: stat,
        key: key,
        days: days,
        tier: tier
    };
    if(errCode){
        task.errMessage = errCode;
        task.endTime = new Date();
    }
    self.taskManager.push(task);
};

TaskControlAction.prototype.pasteObject = function(targetBucket, targetKey, sourceBucket, sourceKey, size, action, stat, errCode) {
    var self = this;
    var uuid = guid();
    var task = {
        id: uuid,
        type: action,
        bucket: targetBucket,
        size: size,
        stat: stat,
        key: targetKey,
        sourceBucket: sourceBucket,
        sourceKey: sourceKey,
        modifyTime: new Date(),
        startTime: new Date()
    };
    if(errCode){
        task.errMessage = errCode;
    }
    self.taskManager.push(task);
}

module.exports = TaskControlAction;