var AWS = require('../core');
var util = require('../util');
var Rest = require('./rest');

function populateBody(req)
{
    var input = req.service.api.operations[req.operation].input;
    var builder = new AWS.XML.Builder();
    var params = req.params;

    var payload = input.payload;
    if (payload)
    {
        var payloadMember = input.members[payload];
        params = params[payload];
        if (params === undefined)
            return;

        if (payloadMember.type === 'structure')
        {
            var rootElement = payloadMember.name;
            if (!(params.LocationConstraint == "us-east-1" &&
                req.operation == "createBucket")) {//Ϊ˼AWSregionͰͰз壬ӰUDS
                req.httpRequest.body = builder.toXML(params, payloadMember, rootElement);
            }
        }
        else
        { // non-xml payload
            req.httpRequest.body = params;
        }
    }
    else {
        if (req.operation != "copyObject" && req.operation != "createMultipartUpload") {//AWSʼûз壬UDSûзĳʼ
            req.httpRequest.body = builder.toXML(params, input, input.name || input.shape || util.string.upperFirst(req.operation) + 'Request');
        }
    }
}

function buildRequest(req)
{
    Rest.buildRequest(req);

    // never send body payload on GET/HEAD
    if ([ 'GET', 'HEAD' ].indexOf(req.httpRequest.method) < 0)
    {
        populateBody(req);
    }
}

function extractError(resp)
{
    Rest.extractError(resp);

    var data = new AWS.XML.Parser().parse(resp.httpResponse.body.toString());
    if (data.Errors)
        data = data.Errors;
    if (data.Error)
        data = data.Error;
    if (data.Code)
    {
        if(data.Code=='XMLParserError'&&(resp.httpResponse.statusCode == 302||resp.httpResponse.statusCode == 407||
			resp.httpResponse.statusCode == 304||resp.httpResponse.statusCode == 404||
			resp.httpResponse.statusCode == 305||resp.httpResponse.statusCode == 400||
			resp.httpResponse.statusCode == 415||resp.httpResponse.statusCode == 502)){
            resp.error = util.error(new Error(), {
                code : resp.httpResponse.statusCode,
                message : data.Message
            });
        }else if(data.Code=='XMLParserError'&&resp.httpResponse.statusCode == 403){
            resp.error = util.error(new Error(), {
                code : 'AccessDenied',
                message : data.Message
            });
        }else{
            resp.error = util.error(new Error(), {
                code : data.Code,
                message : data.Message
            });
        }
    }
    else
    {
        resp.error = util.error(new Error(), {
            code : resp.httpResponse.statusCode,
            message : null
        });
    }
}

function extractData(resp)
{

    Rest.extractData(resp);

    var parser;
    var req = resp.request;
    var body = resp.httpResponse.body;
    var operation = req.service.api.operations[req.operation];
    var output = operation.output;

    var payload = output.payload;
    if (payload)
    {
        var payloadMember = output.members[payload];
        if (payloadMember.isStreaming)
        {
            resp.data[payload] = body;
        }
        else
            if (payloadMember.type === 'structure')
            {
                parser = new AWS.XML.Parser();
                resp.data[payload] = parser.parse(body.toString(), payloadMember);
            }
            else
            {
                resp.data[payload] = body.toString();
            }
    }
    else
        if (body.length > 0)
        {
            parser = new AWS.XML.Parser();
            var data = parser.parse(body.toString(), output);

            util.update(resp.data, data);
        }
}

module.exports = {
    buildRequest : buildRequest,
    extractError : extractError,
    extractData : extractData
};
