var AWS = require('../core');
var fs = require('fs');

AWS.S3.ManagedDownload = AWS.util.inherit({

    constructor : function ManagedDownload(options)
    {
        var self = this;
        AWS.SequentialExecutor.call(self);
        self.size = 0;
        self.localpath=null;

        self.callback = null;
        self.parts = {};
        self.completeMap = {};

        global.logger.info("download:",options.params.Key);
        self.configure(options);
        //等待发送的send请求数，用来控制下载速度
        self.sendQueueSize = 0;
        //控制发送send请求的频率
        self.sendRate = setInterval(function () {
            if(self.sendQueueSize > 0){
                self.sendQueueSize -- ;
                self.fillStream();
            }
        },1000/global.downloadRate);
    },


    configure : function configure(options)
    {
        options = options || {};
        this.partSize = this.minPartSize;
        var total = 0;
        if (options.queueSize)
            this.queueSize = options.queueSize;
        if (options.partSize)
            this.partSize = options.partSize;
        if (options.leavePartsOnError)
            this.leavePartsOnError = true;
        if(options.completeMap){
            this.completeMap=options.completeMap;
            for(var i in options.completeMap){
                total = total + options.completeMap[i]-(i-1)*options.partSize;
            }
            options.totalDownloaded = total;
        }
        if(options.totalDownloaded){
            this.totalDownloadedBytes=options.totalDownloaded;
        }


        this.service = options.service;
        this.bindServiceObject(options.params);



        this.size = this.service.config.params.size;
        this.localpath = this.service.config.params.localpath;
        this.totalBytes = this.size;

    },

    /**
     * @api private
     */
    leavePartsOnError : false,


    queueSize : 4,


    partSize : null,


    minPartSize : 1024 * 1024 * 5,


    maxTotalParts : 10000,


    send : function(callback)
    {
        var self = this;
        self.callback = callback || function(err)
        {

            if (err)
                throw err;
        };


        if (AWS.util.isNode())
        {
            if (self.size <= self.partSize)
                self.sendQueueSize ++;
            else
                for(var i=0;i<self.queueSize;i++) {
                    self.sendQueueSize ++;
                }

        }


    },


    abort : function(callback)
    {
        this.cleanup(AWS.util.error(new Error('Request aborted by user'), {
            code : 'RequestAbortedError',
            retryable : false
        }));
		if(this.fd)
			fs.closeSync(this.fd);
        if(callback){
            callback.call();
        }
    },



    pause : function(callback)
    {
        var self = this;
        if (self.failed)
            return;
        self.failed = true;
        AWS.util.each(self.parts, function(partNumber, part)
        {
            part.removeAllListeners('httpDownloadProgress');
            part.removeAllListeners('complete');
            part.removeAllListeners('httpData');
            part.abort();
        });
        self.clearFd(callback);
    },



    bindServiceObject : function bindServiceObject(params)
    {
        params = params || {};
        var self = this;

        // bind parameters to new service object
        if (!self.service)
        {
            self.service = new AWS.S3({
                params : params
            });
        }
        else
        {
            var config = AWS.util.copy(self.service.config);
            self.service = new self.service.constructor.__super__(config);
            self.service.config.params = AWS.util.merge(self.service.config.params || {}, params);
        }
    },





    isDoneChunking : false,


    partPos : 0,


    totalChunkedBytes : 0,


    totalDownloadedBytes : 0,


    totalDownloaded : 0,


    totalBytes : undefined,


    numParts : 0,


    totalPartNumbers : 0,


    activeParts : 0,


    doneParts : 0,


    parts : null,


    completeMap : {},


    failed : false,


    multipartReq : null,


    partBuffer : null,

    fileTotalSize : 0,

    writeErrNum : 0,

    writeNum : 0,

    fillStream : function fillStream()
    {
        var self = this;

        if (self.activeParts >= self.queueSize)
            return;
        //算出下载的文件位置
        var start = self.totalChunkedBytes
        var end = (self.totalChunkedBytes + self.partSize < self.size) ? self.totalChunkedBytes + self.partSize - 1 : self.size-1
        self.totalChunkedBytes = end + 1;

        //结束
        if(end===self.size-1){
            self.isDoneChunking=true;

        }
        //只有一片
        if (self.size <= self.partSize)
        {
            self.isDoneChunking=true;
            self.totalBytes = self.size;
            if (self.size > 0)
            {
                self.numParts++;
                self.nextChunk(0,self.size-1);
            }
        }
        //大于一片
        else
        {
            if(start<=end){
                self.nextChunk(start,end);

            }
        }

    },


    nextChunk : function nextChunk(start,end)
    {
        var self = this;
        self.activeParts++;
        if (self.failed)
            return null;

        var partNumber = ++self.totalPartNumbers;
        if (self.isDoneChunking && partNumber === 1)
        {
            var position=start;
            if(self.completeMap[partNumber]){
                //已下载部分
                position=self.completeMap[partNumber];
                var downloaSize = position-start;
                self.fileTotalSize+=downloaSize;
                if(self.fileTotalSize===self.size){
                    self.finishSinglePart(null);
                }
            }

            //该段已经下载完成
            if(position>=end) {
            }
            //没有下载或没下载完成
            else{
                var partParams={
                    Range:"bytes="+position+"-"
                };


                var req = self.service.getObject(partParams);
                self.parts[partNumber] = req;
                req._managedDownload = self;

                if(!self.fd){
                    try{
                        var exist=fs.existsSync(self.localpath);
                        var fd;
                        if(exist){
                            fd=fs.openSync(self.localpath, 'r+');
                        }else{
                                fd=fs.openSync(self.localpath, 'w+');
                        }
                        self.fd=fd;
                    }catch(err){
                        global.logger.error('openSync file error:', err);
                        self.clearFd(function () {
                            return self.cleanup(err);
                        });
                    }
                }


                req.position=position;
                req.
                    on('httpDownloadProgress', self.progress).
                    on('httpData', function (data,resp) {
                        if(!self.fd){
                            return;
                        }
						if(resp.request && resp.request.response && resp.request.response.httpResponse && resp.request.response.httpResponse.statusCode > 300){
                                return;
                            }
						self.writeNum++;
                        fs.write(self.fd, data, 0, data.length, this.position,function(err, written, buffer){
							self.writeNum--;
                            if(err){
                                global.logger.error('write data error:', err);
                                if(err.code=='ENOSPC'){
                                    AWS.util.each(self.parts, function(partNumber, part)
                                    {
                                        part.removeAllListeners('httpDownloadProgress');
                                        part.removeAllListeners('complete');
                                        part.removeAllListeners('httpData');
                                        part.abort();
                                    });
                                    self.clearFd(function () {
                                        return self.cleanup(err);
                                    });
                                }
                            }else{
                                self.fileTotalSize+=data.length;
                                if(self.fileTotalSize === self.size){
                                    self.finishSinglePart(null);
                                }
                            }
                            data=null;
                            delete data;
                        });
                        this.position+=data.length;
                        self.completeMap[partNumber]=this.position;
                    }).
                    on('httpError', function (err) {
                        self.clearFd(function () {
                            return self.cleanup(err);
                        });
                    })

                req.send(function(err){
                    if(err){
                        self.clearFd(function () {
                            return self.cleanup(err);
                        });
                    }
                });
                return null;
            }
        }

        else
        {
            self.downloadPart(start,end, partNumber);
        }

    },


    downloadPart : function downloadPart(start, end, partNumber)
    {
        var self = this;


        var position=start;
        if(self.completeMap[partNumber]){
            //已下载部分
            position=self.completeMap[partNumber];
            var downloaSize = position-start;
            self.fileTotalSize+=downloaSize;
            if(self.fileTotalSize===self.size){
                global.logger.info('downloadPart end!');
                self.finishMultiPart();
            }
        }

        //该段已经下载完成
        if(position>end){
            setTimeout(function(){
                chunk=null;

                self.activeParts--;
                self.doneParts++;

                if (self.isDoneChunking && self.doneParts === self.numParts)
                {

                }
                else
                {
                    self.sendQueueSize ++;
                }
            },1000);


        }
        //没有下载
        else{
            var partParams={
                Range:"bytes="+position+"-"+end
            };

            var req = self.service.getObject(partParams);
            self.parts[partNumber] = req;
            req._lastDownloadedBytes = 0;
            req._managedDownload = self;
            req.partNumber=partNumber;

            if(!self.fd){
                try{
                    var exist=fs.existsSync(self.localpath);
                    var fd;
                    if(exist){
                        fd=fs.openSync(self.localpath, 'r+');
                    }else{
                        fd=fs.openSync(self.localpath, 'w+');
                    }
                    self.fd=fd;
                }catch(err){
                    global.logger.error('openSync file error:', err);
                    self.clearFd(function () {
                        return self.cleanup(err);
                    });
                }
            }

            req.position=position;
            req.
                on('httpDownloadProgress', self.progress).
                on('httpData', function (data,resp) {
                    if(!self.fd){
                        return;
                    }
					if(resp.request && resp.request.response && resp.request.response.httpResponse && resp.request.response.httpResponse.statusCode > 300){
				        global.logger.error('httpData statusCode:', resp.request.response.httpResponse.statusCode);
				        return;
                    }
                    self.writeNum++;
                    fs.write(self.fd, data, 0, data.length, this.position,(function(position,self,data,partNumber){
                        return function(err, written, buffer){
                            self.writeNum--;
                            if(err){
                                if(position < self.completeMap[partNumber]){
                                    self.writeErrNum++;
                                    global.logger.info('self.completeMap[partNumber]:',self.completeMap[partNumber],position);
                                    self.completeMap[partNumber] = position;
                                }

                                //如果出现磁盘空间不足的错误，则将任务置为失败状态。
                                if(err.code=='ENOSPC'){
                                    global.logger.error('write file error:', err);
                                    AWS.util.each(self.parts, function(partNumber, part)
                                    {
                                        part.removeAllListeners('httpData');
                                    });
                                    self.clearFd(function () {
                                        return self.cleanup(err);
                                    });
                                }
                            }else{
                                 self.fileTotalSize+=data.length;
                                 if(self.fileTotalSize === self.size){
                                       self.finishMultiPart();
                                 }
                            }
                            data=null;
                            delete data;
                        }
                    })(this.position,self,data,partNumber));
                    this.position+=data.length;
                    self.completeMap[partNumber]=this.position;

                }).
                on('httpError', function (err,data) {
                    //关闭文件
                    console.log('HTTPERROR==================',err);
                    self.clearFd(function () {
                        return self.cleanup(err);
                    });
                })


            req.send(function(err)
            {
                delete self.parts[partNumber];
                self.activeParts--;

                if (err){
                    global.logger.error('send data error:', err);
                    self.clearFd(function () {
                        return self.cleanup(err);
                    });
                }


                self.totalDownloaded+=((end-start)+1);

                self.doneParts++;

                if (self.isDoneChunking && self.numParts === self.doneParts)
                {
                }
                else
                {
                    self.sendQueueSize ++;
                }
            });
        }

    },

    /**
     * @api private
     */
    cleanup : function cleanup(err)
    {
        var self = this;
        if (self.failed)
            return;
        self.failed = true;

        AWS.util.each(self.parts, function(partNumber, part)
        {
            part.removeAllListeners('complete');
            part.abort();
        });
        self.parts = {};
        self.callback(err);
        clearInterval(self.sendRate);
    },


    /**
     * @api private
     */
    finishMultiPart : function finishMultiPart()
    {
        var self = this;
        //关闭文件
        global.logger.info('finishMultiPart:close the fd!');
        if(self.fd){
            fs.closeSync(self.fd);
        }
        var data={};
        data.msg="下载完成";
        clearInterval(this.sendRate);
        self.callback(null, data);
        delete  data;
    },

    /**
     * @api private
     */
    finishSinglePart : function finishSinglePart(err, data)
    {
        var download = this;
        if(download.fd){
            fs.closeSync(download.fd);
        }
        if (err)
            return download.callback(err);
        clearInterval(this.sendRate);
        download.callback(err, data);
    },

    /**
     * @api private
     * 进度回掉函数
     */
    progress : function progress(info)
    {
        var download = this._managedDownload;
        download.totalDownloadedBytes += info.loaded - this._lastDownloadedBytes;
        this._lastDownloadedBytes = info.loaded;
        info = {
            loaded :download.totalDownloadedBytes,
            total : download.totalBytes,
            part : this.partNumber
        };
        download.emit('httpDownloadProgress', [ info ]);
        delete info;
        info = null;
    },
    clearFd:function clearFd(callback){
        var self = this;
        AWS.util.each(self.parts, function(partNumber, part)
        {
            part.removeAllListeners('httpDownloadProgress');
            part.removeAllListeners('complete');
            part.removeAllListeners('httpData');
            part.abort();
        });
        var num = 0;
        var t = setInterval(function(){
	        global.logger.info("Download logger:writeNum",self.writeNum);
            if(self.writeNum===0){
                clearInterval(t);
                if(self.fd){
                    fs.closeSync(self.fd);
                }
                if(callback){
                    callback.call();
                }
            }else if(num>1200){
                clearInterval(t);
                AWS.util.each(self.parts, function(partNumber)
                {
                    self.completeMap[partNumber]=(partNumber-1)*this.partSize;
                });
                global.logger.info("Download logger: Close FD timeout",num);
                if(self.fd){
                    fs.closeSync(self.fd);
                }
                if(callback){
                    callback.call();
                }
            }else{
                num++;
            }
        },100);
    }
});

AWS.util.mixin(AWS.S3.ManagedDownload, AWS.SequentialExecutor);
module.exports = AWS.S3.ManagedDownload;
